"""
This module contains utility functions for writing ddtrace integrations.
"""
from collections import deque
import ipaddress
import os
import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union
from typing import cast

from ddtrace import Pin
from ddtrace import config
from ddtrace.ext import http
from ddtrace.ext import user
from ddtrace.internal import _context
from ddtrace.internal.compat import six
from ddtrace.internal.logger import get_logger
from ddtrace.internal.utils.cache import cached
from ddtrace.internal.utils.formats import asbool
from ddtrace.internal.utils.http import normalize_header_name
from ddtrace.internal.utils.http import redact_url
from ddtrace.internal.utils.http import strip_query_string
import ddtrace.internal.utils.wrappers
from ddtrace.propagation.http import HTTPPropagator
from ddtrace.vendor import wrapt


if TYPE_CHECKING:
    from ddtrace import Span
    from ddtrace import Tracer
    from ddtrace.settings import IntegrationConfig


log = get_logger(__name__)

wrap = wrapt.wrap_function_wrapper
unwrap = ddtrace.internal.utils.wrappers.unwrap
iswrapped = ddtrace.internal.utils.wrappers.iswrapped

REQUEST = "request"
RESPONSE = "response"

# Tag normalization based on: https://docs.datadoghq.com/tagging/#defining-tags
# With the exception of '.' in header names which are replaced with '_' to avoid
# starting a "new object" on the UI.
NORMALIZE_PATTERN = re.compile(r"([^a-z0-9_\-:/]){1}")

# Possible User Agent header.
USER_AGENT_PATTERNS = ("http-user-agent", "user-agent")

IP_PATTERNS = (
    "x-forwarded-for",
    "x-real-ip",
    "client-ip",
    "x-forwarded",
    "x-cluster-client-ip",
    "forwarded-for",
    "forwarded",
    "via",
    "true-client-ip",
)


@cached()
def _normalized_header_name(header_name):
    # type: (str) -> str
    return NORMALIZE_PATTERN.sub("_", normalize_header_name(header_name))


def _get_header_value_case_insensitive(headers, keyname):
    # type: (Mapping[str, str], str) -> Optional[str]
    """
    Get a header in a case insensitive way. This function is meant for frameworks
    like Django < 2.2 that don't store the headers in a case insensitive mapping.
    """
    # just in case we are lucky
    shortcut_value = headers.get(keyname)
    if shortcut_value is not None:
        return shortcut_value

    for key, value in six.iteritems(headers):
        if key.lower().replace("_", "-") == keyname:
            return value

    return None


def _normalize_tag_name(request_or_response, header_name):
    # type: (str, str) -> str
    """
    Given a tag name, e.g. 'Content-Type', returns a corresponding normalized tag name, i.e
    'http.request.headers.content_type'. Rules applied actual header name are:
    - any letter is converted to lowercase
    - any digit is left unchanged
    - any block of any length of different ASCII chars is converted to a single underscore '_'
    :param request_or_response: The context of the headers: request|response
    :param header_name: The header's name
    :type header_name: str
    :rtype: str
    """
    # Looking at:
    #   - http://www.iana.org/assignments/message-headers/message-headers.xhtml
    #   - https://tools.ietf.org/html/rfc6648
    # and for consistency with other language integrations seems safe to assume the following algorithm for header
    # names normalization:
    #   - any letter is converted to lowercase
    #   - any digit is left unchanged
    #   - any block of any length of different ASCII chars is converted to a single underscore '_'
    normalized_name = _normalized_header_name(header_name)
    return "http.{}.headers.{}".format(request_or_response, normalized_name)


def _store_headers(headers, span, integration_config, request_or_response):
    # type: (Dict[str, str], Span, IntegrationConfig, str) -> None
    """
    :param headers: A dict of http headers to be stored in the span
    :type headers: dict or list
    :param span: The Span instance where tags will be stored
    :type span: ddtrace.span.Span
    :param integration_config: An integration specific config object.
    :type integration_config: ddtrace.settings.IntegrationConfig
    """
    if not isinstance(headers, dict):
        try:
            headers = dict(headers)
        except Exception:
            return

    if integration_config is None:
        log.debug("Skipping headers tracing as no integration config was provided")
        return

    for header_name, header_value in headers.items():
        """config._header_tag_name gets an element of the dictionary in config.http._header_tags
        which gets the value from DD_TRACE_HEADER_TAGS environment variable."""
        tag_name = integration_config._header_tag_name(header_name)
        if tag_name is None:
            continue
        # An empty tag defaults to a http.<request or response>.headers.<header name> tag
        span.set_tag(tag_name or _normalize_tag_name(request_or_response, header_name), header_value)


def _get_request_header_user_agent(headers, headers_are_case_sensitive=False):
    # type: (Mapping[str, str], bool) -> str
    """Get user agent from request headers
    :param headers: A dict of http headers to be stored in the span
    :type headers: dict or list
    """
    for key_pattern in USER_AGENT_PATTERNS:
        if not headers_are_case_sensitive:
            user_agent = headers.get(key_pattern)
        else:
            user_agent = _get_header_value_case_insensitive(headers, key_pattern)

        if user_agent:
            return user_agent
    return ""


# Used to cache the last header used for the cache. From the same server/framework
# usually the same header will be used on further requests, so we use this to check
# only it.
_USED_IP_HEADER = ""


def _get_request_header_client_ip(span, headers, peer_ip=None, headers_are_case_sensitive=False):
    # type: (Span, Mapping[str, str], Optional[str], bool) -> str
    global _USED_IP_HEADER

    if asbool(os.getenv("DD_TRACE_CLIENT_IP_HEADER_DISABLED", default=False)):
        return ""

    def get_header_value(key):  # type: (str) -> Optional[str]
        if not headers_are_case_sensitive:
            return headers.get(key)

        return _get_header_value_case_insensitive(headers, key)

    ip_header_value = ""
    user_configured_ip_header = os.getenv("DD_TRACE_CLIENT_IP_HEADER", None)
    if user_configured_ip_header:
        # Used selected the header to use to get the IP
        ip_header_value = headers.get(user_configured_ip_header)
        if not ip_header_value:
            log.debug("DD_TRACE_CLIENT_IP_HEADER configured but '%s' header missing", user_configured_ip_header)
            return ""

        try:
            _ = ipaddress.ip_address(six.text_type(ip_header_value))
        except ValueError:
            log.debug("Invalid IP address from configured %s header: %s", user_configured_ip_header, ip_header_value)
            return ""

    else:
        # No configured IP header, go through the list defined in:
        # https://datadoghq.atlassian.net/wiki/spaces/APS/pages/2118779066/Client+IP+addresses+resolution
        if _USED_IP_HEADER:
            ip_header_value = get_header_value(_USED_IP_HEADER)

        # For some reason request uses other header or the specified header is empty, do the
        # search
        if not ip_header_value:
            for ip_header in IP_PATTERNS:
                tmp_ip_header_value = get_header_value(ip_header)
                if tmp_ip_header_value:
                    ip_header_value = tmp_ip_header_value
                    _USED_IP_HEADER = ip_header
                    break

    # At this point, we have one IP header, check its value and retrieve the first public IP
    private_ip = ""

    if ip_header_value:
        ip_list = ip_header_value.split(",")
        for ip in ip_list:
            ip = ip.strip()
            if not ip:
                continue

            try:
                ip_obj = ipaddress.ip_address(six.text_type(ip))
            except ValueError:
                continue

            if not ip_obj.is_private:  # is_global is Python3+ only
                return ip
            elif not private_ip and not ip_obj.is_loopback:
                private_ip = ip
            # else: it's private, but we already have one: continue in case we find a public one

    # So we have none or maybe one private ip, check the peer ip in case it's public and if not
    # return either the private_ip from the headers (if we have one) or the peer private ip
    if peer_ip:
        try:
            peer_ip_obj = ipaddress.ip_address(six.text_type(peer_ip))
            if not peer_ip_obj.is_loopback:
                # If we have a private_ip and the peer_ip is not public, prefer the one we have
                if private_ip:
                    if peer_ip_obj.is_loopback or peer_ip_obj.is_private:
                        return private_ip
                # peer_ip is public or we dont have a private_ip to we return the private peer_ip
                return peer_ip
        except ValueError:
            pass

    return private_ip


def _store_request_headers(headers, span, integration_config):
    # type: (Dict[str, str], Span, IntegrationConfig) -> None
    """
    Store request headers as a span's tags
    :param headers: All the request's http headers, will be filtered through the whitelist
    :type headers: dict or list
    :param span: The Span instance where tags will be stored
    :type span: ddtrace.Span
    :param integration_config: An integration specific config object.
    :type integration_config: ddtrace.settings.IntegrationConfig
    """
    _store_headers(headers, span, integration_config, REQUEST)


def _store_response_headers(headers, span, integration_config):
    # type: (Dict[str, str], Span, IntegrationConfig) -> None
    """
    Store response headers as a span's tags
    :param headers: All the response's http headers, will be filtered through the whitelist
    :type headers: dict or list
    :param span: The Span instance where tags will be stored
    :type span: ddtrace.Span
    :param integration_config: An integration specific config object.
    :type integration_config: ddtrace.settings.IntegrationConfig
    """
    _store_headers(headers, span, integration_config, RESPONSE)


def with_traced_module(func):
    """Helper for providing tracing essentials (module and pin) for tracing
    wrappers.

    This helper enables tracing wrappers to dynamically be disabled when the
    corresponding pin is disabled.

    Usage::

        @with_traced_module
        def my_traced_wrapper(django, pin, func, instance, args, kwargs):
            # Do tracing stuff
            pass

        def patch():
            import django
            wrap(django.somefunc, my_traced_wrapper(django))
    """

    def with_mod(mod):
        def wrapper(wrapped, instance, args, kwargs):
            pin = Pin._find(instance, mod)
            if pin and not pin.enabled():
                return wrapped(*args, **kwargs)
            elif not pin:
                log.debug("Pin not found for traced method %r", wrapped)
                return wrapped(*args, **kwargs)
            return func(mod, pin, wrapped, instance, args, kwargs)

        return wrapper

    return with_mod


def distributed_tracing_enabled(int_config, default=False):
    # type: (IntegrationConfig, bool) -> bool
    """Returns whether distributed tracing is enabled for this integration config"""
    if "distributed_tracing_enabled" in int_config and int_config.distributed_tracing_enabled is not None:
        return int_config.distributed_tracing_enabled
    elif "distributed_tracing" in int_config and int_config.distributed_tracing is not None:
        return int_config.distributed_tracing
    return default


def int_service(pin, int_config, default=None):
    # type: (Optional[Pin], IntegrationConfig, Optional[str]) -> Optional[str]
    """Returns the service name for an integration which is internal
    to the application. Internal meaning that the work belongs to the
    user's application. Eg. Web framework, sqlalchemy, web servers.

    For internal integrations we prioritize overrides, then global defaults and
    lastly the default provided by the integration.
    """
    # Pin has top priority since it is user defined in code
    if pin is not None and pin.service:
        return pin.service

    # Config is next since it is also configured via code
    # Note that both service and service_name are used by
    # integrations.
    if "service" in int_config and int_config.service is not None:
        return cast(str, int_config.service)
    if "service_name" in int_config and int_config.service_name is not None:
        return cast(str, int_config.service_name)

    global_service = int_config.global_config._get_service()
    if global_service:
        return cast(str, global_service)

    if "_default_service" in int_config and int_config._default_service is not None:
        return cast(str, int_config._default_service)

    return default


def ext_service(pin, int_config, default=None):
    # type: (Optional[Pin], IntegrationConfig, Optional[str]) -> Optional[str]
    """Returns the service name for an integration which is external
    to the application. External meaning that the integration generates
    spans wrapping code that is outside the scope of the user's application. Eg. A database, RPC, cache, etc.
    """
    if pin is not None and pin.service:
        return pin.service

    if "service" in int_config and int_config.service is not None:
        return cast(str, int_config.service)
    if "service_name" in int_config and int_config.service_name is not None:
        return cast(str, int_config.service_name)

    if "_default_service" in int_config and int_config._default_service is not None:
        return cast(str, int_config._default_service)

    # A default is required since it's an external service.
    return default


def set_http_meta(
    span,  # type: Span
    integration_config,  # type: IntegrationConfig
    method=None,  # type: Optional[str]
    url=None,  # type: Optional[str]
    status_code=None,  # type: Optional[Union[int, str]]
    status_msg=None,  # type: Optional[str]
    query=None,  # type: Optional[str]
    parsed_query=None,  # type: Optional[Mapping[str, str]]
    request_headers=None,  # type: Optional[Mapping[str, str]]
    response_headers=None,  # type: Optional[Mapping[str, str]]
    retries_remain=None,  # type: Optional[Union[int, str]]
    raw_uri=None,  # type: Optional[str]
    request_cookies=None,  # type: Optional[Dict[str, str]]
    request_path_params=None,  # type: Optional[Dict[str, str]]
    request_body=None,  # type: Optional[Union[str, Dict[str, List[str]]]]
    peer_ip=None,  # type: Optional[str]
    headers_are_case_sensitive=False,  # type: bool
):
    # type: (...) -> None
    """
    Set HTTP metas on the span

    :param method: the HTTP method
    :param url: the HTTP URL
    :param status_code: the HTTP status code
    :param status_msg: the HTTP status message
    :param query: the HTTP query part of the URI as a string
    :param parsed_query: the HTTP query part of the URI as parsed by the framework and forwarded to the user code
    :param request_headers: the HTTP request headers
    :param response_headers: the HTTP response headers
    :param raw_uri: the full raw HTTP URI (including ports and query)
    :param request_cookies: the HTTP request cookies as a dict
    :param request_path_params: the parameters of the HTTP URL as set by the framework: /posts/<id:int> would give us
         { "id": <int_value> }
    """
    if method is not None:
        span._set_str_tag(http.METHOD, method)

    if url is not None:
        if integration_config.trace_query_string and not config.global_trace_query_string_disabled:
            span._set_str_tag(http.URL, redact_url(url, config._obfuscation_query_string_pattern, query))
        else:
            span._set_str_tag(http.URL, strip_query_string(url))

    if status_code is not None:
        try:
            int_status_code = int(status_code)
        except (TypeError, ValueError):
            log.debug("failed to convert http status code %r to int", status_code)
        else:
            span._set_str_tag(http.STATUS_CODE, str(status_code))
            if config.http_server.is_error_code(int_status_code):
                span.error = 1

    if status_msg is not None:
        span._set_str_tag(http.STATUS_MSG, status_msg)

    if query is not None and integration_config.trace_query_string:
        span._set_str_tag(http.QUERY_STRING, query)

    ip = None
    if request_headers:
        user_agent = _get_request_header_user_agent(request_headers, headers_are_case_sensitive)
        if user_agent:
            span.set_tag(http.USER_AGENT, user_agent)

        if integration_config.is_header_tracing_configured:
            """We should store both http.<request_or_response>.headers.<header_name> and http.<key>. The last one
            is the DD standardized tag for user-agent"""
            _store_request_headers(dict(request_headers), span, integration_config)

        ip = _get_request_header_client_ip(span, request_headers, peer_ip, headers_are_case_sensitive)
        if ip:
            span.set_tag(http.CLIENT_IP, ip)
            span.set_tag("network.client.ip", ip)

    if response_headers is not None and integration_config.is_header_tracing_configured:
        _store_response_headers(dict(response_headers), span, integration_config)

    if retries_remain is not None:
        span._set_str_tag(http.RETRIES_REMAIN, str(retries_remain))

    if config._appsec_enabled:
        status_code = str(status_code) if status_code is not None else None

        _context.set_items(
            {
                k: v
                for k, v in [
                    ("http.request.uri", raw_uri),
                    ("http.request.method", method),
                    ("http.request.cookies", request_cookies),
                    ("http.request.query", parsed_query),
                    ("http.request.headers", request_headers),
                    ("http.response.headers", response_headers),
                    ("http.response.status", status_code),
                    ("http.request.path_params", request_path_params),
                    ("http.request.body", request_body),
                    ("http.request.remote_ip", ip),
                ]
                if v is not None
            },
            span=span,
        )


def activate_distributed_headers(tracer, int_config=None, request_headers=None, override=None):
    # type: (Tracer, Optional[IntegrationConfig], Optional[Dict[str, str]], Optional[bool]) -> None
    """
    Helper for activating a distributed trace headers' context if enabled in integration config.
    int_config will be used to check if distributed trace headers context will be activated, but
    override will override whatever value is set in int_config if passed any value other than None.
    """
    if override is False:
        return None

    if override or (int_config and distributed_tracing_enabled(int_config)):
        context = HTTPPropagator.extract(request_headers)

        # Only need to activate the new context if something was propagated
        if not context.trace_id:
            return None

        # Do not reactivate a context with the same trace id
        # DEV: An example could be nested web frameworks, when one layer already
        #      parsed request headers and activated them.
        #
        # Example::
        #
        #     app = Flask(__name__)  # Traced via Flask instrumentation
        #     app = DDWSGIMiddleware(app)  # Extra layer on top for WSGI
        current_context = tracer.current_trace_context()
        if current_context and current_context.trace_id == context.trace_id:
            log.debug(
                "will not activate extracted Context(trace_id=%r, span_id=%r), a context with that trace id is already active",  # noqa: E501
                context.trace_id,
                context.span_id,
            )
            return None

        # We have parsed a trace id from headers, and we do not already
        # have a context with the same trace id active
        tracer.context_provider.activate(context)


def _flatten(
    obj,  # type: Any
    sep=".",  # type: str
    prefix="",  # type: str
    exclude_policy=None,  # type: Optional[Callable[[str], bool]]
):
    # type: (...) -> Generator[Tuple[str, Any], None, None]
    s = deque()  # type: ignore
    s.append((prefix, obj))
    while s:
        p, v = s.pop()
        if exclude_policy is not None and exclude_policy(p):
            continue
        if isinstance(v, dict):
            s.extend((sep.join((p, k)) if p else k, v) for k, v in v.items())
        else:
            yield p, v


def set_flattened_tags(
    span,  # type: Span
    items,  # type: Iterator[Tuple[str, Any]]
    sep=".",  # type: str
    exclude_policy=None,  # type: Optional[Callable[[str], bool]]
    processor=None,  # type: Optional[Callable[[Any], Any]]
):
    # type: (...) -> None
    for prefix, value in items:
        for tag, v in _flatten(value, sep, prefix, exclude_policy):
            span.set_tag(tag, processor(v) if processor is not None else v)


def set_user(tracer, user_id, name=None, email=None, scope=None, role=None, session_id=None, propagate=False):
    # type: (Tracer, str, Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], bool) -> None
    """Set user tags.
    https://docs.datadoghq.com/logs/log_configuration/attributes_naming_convention/#user-related-attributes
    https://docs.datadoghq.com/security_platform/application_security/setup_and_configure/?tab=set_tag&code-lang=python
    """
    span = tracer.current_root_span()
    if span:
        # Required unique identifier of the user
        span.set_tag(user.ID, user_id)
        if propagate:
            span.context.dd_user_id = user_id

        # All other fields are optional
        if name:
            span.set_tag(user.NAME, name)
        if email:
            span.set_tag(user.EMAIL, email)
        if scope:
            span.set_tag(user.SCOPE, scope)
        if role:
            span.set_tag(user.ROLE, role)
        if session_id:
            span.set_tag(user.SESSION_ID, session_id)
    else:
        log.warning(
            "No root span in the current execution. Skipping set_user tags. "
            "See https://docs.datadoghq.com/security_platform/application_security/setup_and_configure/"
            "?tab=set_user&code-lang=python for more information.",
        )



# -*- coding: utf-8 -*-
"""
Simple test to test tracing. Inspired by CherryPy Tutorials:
https://github.com/cherrypy/cherrypy/blob/master/cherrypy/tutorial/tut01_helloworld.py
"""

import logging
import os.path
import sys

# Import CherryPy global namespace
import cherrypy


logger = logging.getLogger()
logger.level = logging.DEBUG


class HandleMe(Exception):
    pass


class TestError(Exception):
    pass


if sys.version_info[0] < 3:
    UNICODE_ENDPOINT = u"üŋïĉóđē".encode("utf-8")
else:
    UNICODE_ENDPOINT = u"üŋïĉóđē"


class TestApp:
    """Sample request handler class."""

    def __init__(self):
        self.dispatch = TestDispatch()

    @cherrypy.expose
    def index(self):
        return "Hello world!"

    @cherrypy.expose(["aliases", "other"])
    def alias(self):
        return "alias"

    @cherrypy.expose
    def child(self):
        with cherrypy.tools.tracer._tracer.trace("child") as span:
            span.set_tag("a", "b")
            return "child"

    @cherrypy.expose
    def error(self):
        raise TestError()

    @cherrypy.expose
    def handleme(self):
        with cherrypy.HTTPError.handle(HandleMe, status=418, message="handled"):
            raise HandleMe()

    @cherrypy.expose
    def fatal(self):
        1 / 0

    @cherrypy.expose(UNICODE_ENDPOINT)
    def unicode(self):
        logger.info("In the /unicode resource")
        return u"üŋïĉóđē".encode("utf-8")

    @cherrypy.expose
    def custom_span(self):
        span = cherrypy.tools.tracer._tracer.current_span()
        assert span
        span.resource = "overridden"
        return "hiya"

    @cherrypy.expose
    def response_headers(self):
        cherrypy.response.headers["my-response-header"] = "my_response_value"
        return "Hello CherryPy"


@cherrypy.popargs("test_value")
class TestDispatch:
    @cherrypy.expose
    def index(self, test_value):
        return "dispatch with {test_value}".format(test_value=test_value)


testconf = os.path.join(os.path.dirname(__file__), "test.conf")

if __name__ == "__main__":
    # CherryPy always starts with app.root when trying to map request URIs
    # to objects, so we need to mount a request handler root. A request
    # to '/' will be mapped to HelloWorld().index().
    cherrypy.quickstart(TestApp(), config=testconf)



import pytest

from ddtrace import Tracer
from ddtrace.constants import MANUAL_DROP_KEY
from ddtrace.propagation.http import HTTPPropagator
from tests.utils import override_global_config

from .test_integration import AGENT_VERSION


pytestmark = pytest.mark.skipif(AGENT_VERSION != "testagent", reason="Tests only compatible with a testagent")


@pytest.fixture(
    params=[
        dict(global_config=dict()),
        dict(
            global_config=dict(_x_datadog_tags_max_length="0", _x_datadog_tags_enabled=False),
        ),
        dict(global_config=dict(), partial_flush_enabled=True, partial_flush_min_spans=2),
    ]
)
def tracer(request):
    global_config = request.param.get("global_config", dict())
    partial_flush_enabled = request.param.get("partial_flush_enabled")
    partial_flush_min_spans = request.param.get("partial_flush_min_spans")
    with override_global_config(global_config):
        tracer = Tracer()
        kwargs = dict()
        if partial_flush_enabled:
            kwargs["partial_flush_enabled"] = partial_flush_enabled
        if partial_flush_min_spans:
            kwargs["partial_flush_min_spans"] = partial_flush_min_spans
        tracer.configure(**kwargs)
        yield tracer
        tracer.shutdown()


@pytest.mark.snapshot()
def test_trace_tags_multispan(tracer):
    # type: (Tracer) -> None
    headers = {
        "x-datadog-trace-id": "1234",
        "x-datadog-parent-id": "5678",
        "x-datadog-sampling-priority": "1",
        "x-datadog-tags": "_dd.p.dm=-1,_dd.p.test=value,any=tag",
    }
    context = HTTPPropagator.extract(headers)
    # DEV: Trace consists of a simple p->c1 case where c1 is finished before p.
    # But the trace also includes p->c2->gc where c2 and p are finished before
    # gc is finished.
    p = tracer.start_span("p", child_of=context)
    c1 = tracer.start_span("c1", child_of=p)
    c1.finish()
    c2 = tracer.start_span("c2", child_of=p)
    gc = tracer.start_span("gc", child_of=c2)
    c2.finish()
    # If partial flushing had been enabled, this will send [c1, c2] as a trace chunk to the agent.
    tracer.flush()
    p.finish()
    gc.finish()


@pytest.fixture
def downstream_tracer():
    tracer = Tracer()
    yield tracer
    tracer.shutdown()


@pytest.mark.snapshot()
def test_sampling_decision_downstream(downstream_tracer):
    headers = {
        "x-datadog-trace-id": "1234",
        "x-datadog-parent-id": "5678",
        "x-datadog-sampling-priority": "1",
        "x-datadog-tags": "_dd.p.dm=-1",
    }
    context = HTTPPropagator.extract(headers)
    downstream_tracer.context_provider.activate(context)

    with downstream_tracer.trace("p", service="downstream") as span:
        span.set_tag(MANUAL_DROP_KEY)



import warnings

import pytest


def test_not_deprecated():
    import ddtrace

    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter("always")

        assert ddtrace.constants.ENV_KEY
        assert len(ws) == 0


def test_invalid():
    with pytest.raises(ImportError):
        from ddtrace.constants import INVALID_CONSTANT  # noqa



# -*- coding: utf-8 -*-
from collections import OrderedDict

import pytest

from ddtrace.internal._tagset import TagsetDecodeError
from ddtrace.internal._tagset import TagsetEncodeError
from ddtrace.internal._tagset import TagsetMaxSizeEncodeError
from ddtrace.internal._tagset import decode_tagset_string
from ddtrace.internal._tagset import encode_tagset_values
from ddtrace.internal.compat import ensure_str


@pytest.mark.parametrize(
    "header,expected",
    [
        # Normal single tag pair
        ("key=value", {"key": "value"}),
        ("_key=value", {"_key": "value"}),
        # Starting with a digit
        ("1key=value", {"1key": "value"}),
        # Only digits
        ("12345=678910", {"12345": "678910"}),
        # Single character key and value
        ("a=b", {"a": "b"}),
        # Empty value
        ("", {}),
        # Extra trailing comma
        ("key=value,", {"key": "value"}),
        # Values can have spaces
        ("key=value can have spaces", {"key": "value can have spaces"}),
        # Values can have equals
        ("key=value=value", {"key": "value=value"}),
        # Remove leading/trailing spaces from values
        ("key= value can have spaces ", {"key": "value can have spaces"}),
        # Non-space whitespace characters are allowed
        # tenant@vendor format key from tracestate
        ("tenant@vendor=value", {"tenant@vendor": "value"}),
        # Multiple values
        ("a=1,b=2,c=3", {"a": "1", "b": "2", "c": "3"}),
        # Realistic example
        (
            "_dd.p.upstream_services=bWNudWx0eS13ZWI|0|1;dHJhY2Utc3RhdHMtcXVlcnk|2|4,_dd.p.hello=world",
            {"_dd.p.upstream_services": "bWNudWx0eS13ZWI|0|1;dHJhY2Utc3RhdHMtcXVlcnk|2|4", "_dd.p.hello": "world"},
        ),
        (
            "_dd.p.upstream_services=bWNudWx0eS13ZWI===|0|1;dHJhY2Utc3RhdHMtcXVlcnk===|2|4,_dd.p.hello=world",
            {
                "_dd.p.upstream_services": "bWNudWx0eS13ZWI===|0|1;dHJhY2Utc3RhdHMtcXVlcnk===|2|4",
                "_dd.p.hello": "world",
            },
        ),
    ],
)
def test_decode_tagset_string(header, expected):
    """Test that provided header value is parsed as expected"""
    assert expected == decode_tagset_string(header)


@pytest.mark.parametrize(
    "header",
    [
        "key",
        "key=",
        "key=,",
        "=",
        ",",
        ",=,",
        ",=value",
        "=value",
        # Extra leading comma
        ",key=value",
        "key=value,=value",
        "key=value,value",
        # Spaces are not allowed in keys
        "key with spaces=value",
        # Non-space whitespace characters are not allowed in key or value
        "key=value\r\n",
        "key\t=value\r\n",
    ],
)
def test_decode_tagset_string_malformed(header):
    """Test that the provided malformed header values raise an exception"""
    with pytest.raises(TagsetDecodeError):
        decode_tagset_string(header)


@pytest.mark.parametrize(
    "values,expected",
    [
        # No values
        ({}, ""),
        # Single key/value
        ({"key": "value"}, "key=value"),
        # Allow equals in values
        ({"key": "value=with=equals"}, "key=value=with=equals"),
        # Multiple key/values
        # DEV: Use OrderedDict to ensure consistent iteration for encoding
        (OrderedDict([("a", "1"), ("b", "2"), ("c", "3")]), "a=1,b=2,c=3"),
        # Realistic example
        (
            {"_dd.p.upstream_services": "Z3JwYy1jbGllbnQ=|1|0|1.0000"},
            "_dd.p.upstream_services=Z3JwYy1jbGllbnQ=|1|0|1.0000",
        ),
    ],
)
def test_encode_tagset_values(values, expected):
    """Test that we can properly encode data into the expected format"""
    header = encode_tagset_values(values)
    assert expected == header

    # Ensure what we generate also parses correctly
    assert values == decode_tagset_string(header)


def test_encode_tagset_values_strip_spaces():
    """Test that leading and trailing spaces are stripped from keys and values"""
    # Leading/trailing spaces are striped
    values = {" key ": " value "}
    res = encode_tagset_values(values)
    assert "key=value" == res
    assert {"key": "value"} == decode_tagset_string(res)


@pytest.mark.parametrize(
    "values",
    [
        # disallow key with spaces
        {"key with spaces": "value"},
        # disallow commas
        {"key,with,commas": "value"},
        {"key": "value,with,commas"},
        # disallow equals
        {"key=with=equals": "value"},
        # Empty key or value
        {"": "value"},
        {"key": ""},
        # Unicode
        {ensure_str(u"☺️"): "value"},
        {"key": ensure_str(u"☺️")},
    ],
)
def test_encode_tagset_values_malformed(values):
    """Test that invalid data raises a TagsetEncodeError"""
    with pytest.raises(TagsetEncodeError):
        encode_tagset_values(values)


def test_encode_tagset_values_max_size():
    """Test that exceeding the max size raises an exception"""
    # DEV: Use OrderedDict to ensure consistent iteration for encoding
    values = OrderedDict(
        [
            # Short values we know will pack into the final result
            ("a", "1"),
            ("b", "2"),
            ("somereallylongkey", "somereallyreallylongvalue"),
        ]
    )
    with pytest.raises(TagsetMaxSizeEncodeError) as ex_info:
        encode_tagset_values(values, max_size=10)

    ex = ex_info.value
    assert ex.values == values
    assert ex.max_size == 10
    assert ex.current_results == "a=1,b=2"


def test_encode_tagset_values_invalid_type():
    """
    encode_tagset_values accepts `values` as an `object` instead of `dict`
    so we can allow subclasses like `collections.OrderedDict` ensure that
    we properly raise a `TypeError` when a non-dict is passed
    """
    # Allowed
    encode_tagset_values({})
    encode_tagset_values(OrderedDict())

    # Not allowed, AttributeError
    for values in (None, True, 10, object(), []):
        with pytest.raises(AttributeError):
            encode_tagset_values(values)



"""
The ``jinja2`` integration traces templates loading, compilation and rendering.
Auto instrumentation is available using the ``patch``. The following is an example::

    from ddtrace import patch
    from jinja2 import Environment, FileSystemLoader

    patch(jinja2=True)

    env = Environment(
        loader=FileSystemLoader("templates")
    )
    template = env.get_template('mytemplate.html')


The library can be configured globally and per instance, using the Configuration API::

    from ddtrace import config

    # Change service name globally
    config.jinja2['service_name'] = 'jinja-templates'

    # change the service name only for this environment
    cfg = config.get_from(env)
    cfg['service_name'] = 'jinja-templates'

By default, the service name is set to None, so it is inherited from the parent span.
If there is no parent span and the service name is not overridden the agent will drop the traces.
"""
from ...internal.utils.importlib import require_modules


required_modules = ["jinja2"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch
        from .patch import unpatch

        __all__ = [
            "patch",
            "unpatch",
        ]



"""
The ``urllib3`` integration instruments tracing on http calls with optional
support for distributed tracing across services the client communicates with.


Enabling
~~~~~~~~

The ``urllib3`` integration is not enabled by default. Use ``patch_all()``
with the environment variable ``DD_TRACE_URLLIB3_ENABLED`` set, or call
:func:`patch()<ddtrace.patch>` with the ``urllib3`` argument set to ``True`` to manually
enable the integration, before importing and using ``urllib3``::

    from ddtrace import patch
    patch(urllib3=True)

    # use urllib3 like usual


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.urllib3['service']

   The service name reported by default for urllib3 client instances.

   This option can also be set with the ``DD_URLLIB3_SERVICE`` environment
   variable.

   Default: ``"urllib3"``


.. py:data:: ddtrace.config.urllib3['distributed_tracing']

   Whether or not to parse distributed tracing headers.

   Default: ``True``


.. py:data:: ddtrace.config.urllib3['trace_query_string']

   Whether or not to include the query string as a tag.

   Default: ``False``


.. py:data:: ddtrace.config.urllib3['split_by_domain']

   Whether or not to use the domain name of requests as the service name.

   Default: ``False``
"""
from ...internal.utils.importlib import require_modules
from .patch import patch
from .patch import unpatch


required_modules = ["urllib3"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:

        __all__ = [
            "patch",
            "unpatch",
        ]



# -*- encoding: utf-8 -*-
import os
import re
import signal
import sys
import tempfile
import time

import pytest

from tests.contrib.uwsgi import run_uwsgi

from . import utils


# uwsgi does not support Python 3.10 yet
# uwsgi is not available on Windows
if sys.version_info[:2] >= (3, 10) or sys.platform == "win32":
    pytestmark = pytest.mark.skip


uwsgi_app = os.path.join(os.path.dirname(__file__), "uwsgi-app.py")


@pytest.fixture
def uwsgi(monkeypatch):
    # Do not ignore profiler so we have samples in the output pprof
    monkeypatch.setenv("DD_PROFILING_IGNORE_PROFILER", "0")
    # Do not use pytest tmpdir fixtures which generate directories longer than allowed for a socket file name
    socket_name = tempfile.mktemp()
    cmd = ["uwsgi", "--need-app", "--die-on-term", "--socket", socket_name, "--wsgi-file", uwsgi_app]

    try:
        yield run_uwsgi(cmd)
    finally:
        os.unlink(socket_name)


def test_uwsgi_threads_disabled(uwsgi):
    proc = uwsgi()
    stdout, _ = proc.communicate()
    assert proc.wait() != 0
    assert b"ddtrace.internal.uwsgi.uWSGIConfigError: enable-threads option must be set to true" in stdout


def test_uwsgi_threads_enabled(uwsgi, tmp_path, monkeypatch):
    filename = str(tmp_path / "uwsgi.pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    proc = uwsgi("--enable-threads")
    worker_pids = _get_worker_pids(proc.stdout, 1)
    # Give some time to the process to actually startup
    time.sleep(3)
    proc.terminate()
    assert proc.wait() == 30
    for pid in worker_pids:
        utils.check_pprof_file("%s.%d.1" % (filename, pid))


def test_uwsgi_threads_processes_no_master(uwsgi, monkeypatch):
    proc = uwsgi("--enable-threads", "--processes", "2")
    stdout, _ = proc.communicate()
    assert (
        b"ddtrace.internal.uwsgi.uWSGIConfigError: master option must be enabled when multiple processes are used"
        in stdout
    )


def _get_worker_pids(stdout, num_worker, num_app_started=1):
    worker_pids = []
    started = 0
    while True:
        line = stdout.readline()
        if line == b"":
            break
        elif b"WSGI app 0 (mountpoint='') ready" in line:
            started += 1
        else:
            m = re.match(r"^spawned uWSGI worker \d+ .*\(pid: (\d+),", line.decode())
            if m:
                worker_pids.append(int(m.group(1)))

        if len(worker_pids) == num_worker and num_app_started == started:
            break

    return worker_pids


def test_uwsgi_threads_processes_master(uwsgi, tmp_path, monkeypatch):
    filename = str(tmp_path / "uwsgi.pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    proc = uwsgi("--enable-threads", "--master", "--processes", "2")
    worker_pids = _get_worker_pids(proc.stdout, 2)
    # Give some time to child to actually startup
    time.sleep(3)
    proc.terminate()
    assert proc.wait() == 0
    for pid in worker_pids:
        utils.check_pprof_file("%s.%d.1" % (filename, pid))


def test_uwsgi_threads_processes_master_lazy_apps(uwsgi, tmp_path, monkeypatch):
    filename = str(tmp_path / "uwsgi.pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    proc = uwsgi("--enable-threads", "--master", "--processes", "2", "--lazy-apps")
    worker_pids = _get_worker_pids(proc.stdout, 2, 2)
    # Give some time to child to actually startup
    time.sleep(3)
    proc.terminate()
    assert proc.wait() == 0
    for pid in worker_pids:
        utils.check_pprof_file("%s.%d.1" % (filename, pid))


# For whatever reason this crashes easily on Python 2.7 with a segfault, and hangs on Python before 3.7.
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(uwsgi_backtrace+0x35) [0x561586ca5a85]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(uwsgi_segfault+0x23) [0x561586ca5e33]
# /lib/x86_64-linux-gnu/libc.so.6(+0x33060) [0x7f8325888060]
# /root/project/.tox/py27-profile-minreqs-gevent/lib/python2.7/site-packages/
#         google/protobuf/pyext/_message.so(+0x8d450) [0x7f831ed20450]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(+0x14ddb7) [0x561586d36db7]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(PyDict_SetItem+0x67) [0x561586d38327]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(_PyModule_Clear+0x14e) [0x561586d3d21e]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(PyImport_Cleanup+0x380) [0x561586daf440]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(+0x1db008) [0x561586dc4008]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(uwsgi_plugins_atexit+0x81) [0x561586ca2b51]
# /lib/x86_64-linux-gnu/libc.so.6(+0x35940) [0x7f832588a940]
# /lib/x86_64-linux-gnu/libc.so.6(+0x3599a) [0x7f832588a99a]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(+0x7172f) [0x561586c5a72f]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(end_me+0x25) [0x561586ca2b95]
# /lib/x86_64-linux-gnu/libpthread.so.0(+0x110e0) [0x7f83270a50e0]
# /lib/x86_64-linux-gnu/libc.so.6(epoll_wait+0x33) [0x7f832593e303]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(event_queue_wait+0x25) [0x561586c98d55]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(wsgi_req_accept+0x13a) [0x561586c57eba]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(simple_loop_run+0xb6) [0x561586ca1916]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(simple_loop+0x10) [0x561586ca1740]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(uwsgi_ignition+0x1b3) [0x561586ca60a3]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(uwsgi_worker_run+0x28d) [0x561586caaa6d]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(+0xc206c) [0x561586cab06c]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(+0x6e24e) [0x561586c5724e]
# /lib/x86_64-linux-gnu/libc.so.6(__libc_start_main+0xf1) [0x7f83258752e1]
# /root/project/.tox/py27-profile-minreqs-gevent/bin/uwsgi(_start+0x2a) [0x561586c5727a]
@pytest.mark.skipif(
    not (sys.version_info[0] >= 3 and sys.version_info[1] >= 7), reason="this test crashes on old Python versions"
)
def test_uwsgi_threads_processes_no_master_lazy_apps(uwsgi, tmp_path, monkeypatch):
    filename = str(tmp_path / "uwsgi.pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    proc = uwsgi("--enable-threads", "--processes", "2", "--lazy-apps")
    worker_pids = _get_worker_pids(proc.stdout, 2, 2)
    # Give some time to child to actually startup
    time.sleep(3)
    # The processes are started without a master/parent so killing one does not kill the other:
    # Kill them all and wait until they die.
    for pid in worker_pids:
        os.kill(pid, signal.SIGTERM)
    # The first worker is our child, we can wait for it "normally"
    os.waitpid(worker_pids[0], 0)
    # The other ones are grandchildren, we can't wait for it with `waitpid`
    for pid in worker_pids[1:]:
        # Wait for the uwsgi workers to all die
        while True:
            try:
                os.kill(pid, 0)
            except OSError:
                break
    for pid in worker_pids:
        utils.check_pprof_file("%s.%d.1" % (filename, pid))



#!/usr/bin/env python
# coding: utf-8

from msgpack import unpackb

from ddtrace.internal._encoding import packb


def test_unpack_bytearray():
    buf = bytearray(packb(["foo", "bar"]))
    obj = unpackb(buf)
    assert ["foo", "bar"] == obj


def test_unpack_memoryview():
    buf = bytearray(packb(["foo", "bar"]))
    view = memoryview(buf)
    obj = unpackb(view)
    assert ["foo", "bar"] == obj



import json
import os
import sys

import mock
import pytest

from ddtrace import Pin
from ddtrace.constants import SAMPLING_PRIORITY_KEY
from ddtrace.contrib.pytest.constants import XFAIL_REASON
from ddtrace.contrib.pytest.plugin import _extract_repository_name
from ddtrace.ext import ci
from ddtrace.ext import test
from tests.utils import TracerTestCase


class TestPytest(TracerTestCase):
    @pytest.fixture(autouse=True)
    def fixtures(self, testdir, monkeypatch):
        self.testdir = testdir
        self.monkeypatch = monkeypatch

    def inline_run(self, *args):
        """Execute test script with test tracer."""

        class PinTracer:
            @staticmethod
            def pytest_configure(config):
                if Pin.get_from(config) is not None:
                    Pin.override(config, tracer=self.tracer)

        return self.testdir.inline_run(*args, plugins=[PinTracer()])

    def subprocess_run(self, *args):
        """Execute test script with test tracer."""
        return self.testdir.runpytest_subprocess(*args)

    @pytest.mark.skipif(sys.version_info[0] == 2, reason="Triggers a bug with coverage, sqlite and Python 2")
    def test_patch_all(self):
        """Test with --ddtrace-patch-all."""
        py_file = self.testdir.makepyfile(
            """
            import ddtrace

            def test_patched_all():
                assert ddtrace._monkey._PATCHED_MODULES
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace-patch-all", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 0

    @pytest.mark.skipif(sys.version_info[0] == 2, reason="Triggers a bug with coverage, sqlite and Python 2")
    def test_patch_all_init(self):
        """Test with ddtrace-patch-all via ini."""
        self.testdir.makefile(".ini", pytest="[pytest]\nddtrace-patch-all=1\n")
        py_file = self.testdir.makepyfile(
            """
            import ddtrace

            def test_patched_all():
                assert ddtrace._monkey._PATCHED_MODULES
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run(file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 0

    def test_disabled(self):
        """Test without --ddtrace."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            def test_no_trace(ddspan):
                assert ddspan is None
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run(file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 0

    def test_ini(self):
        """Test ini config."""
        self.testdir.makefile(".ini", pytest="[pytest]\nddtrace=1\n")
        py_file = self.testdir.makepyfile(
            """
            import pytest

            def test_ini(ddspan):
                assert ddspan is not None
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run(file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1

    def test_parameterize_case(self):
        """Test parametrize case with simple objects."""
        py_file = self.testdir.makepyfile(
            """
            import pytest


            @pytest.mark.parametrize('item', [1, 2, 3, 4, pytest.param([1, 2, 3], marks=pytest.mark.skip)])
            class Test1(object):
                def test_1(self, item):
                    assert item in {1, 2, 3}
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=3, failed=1, skipped=1)
        spans = self.pop_spans()

        expected_params = [1, 2, 3, 4, [1, 2, 3]]
        assert len(spans) == 5
        for i in range(len(expected_params)):
            assert json.loads(spans[i].get_tag(test.PARAMETERS)) == {
                "arguments": {"item": str(expected_params[i])},
                "metadata": {},
            }

    def test_parameterize_case_complex_objects(self):
        """Test parametrize case with complex objects."""
        py_file = self.testdir.makepyfile(
            """
            from mock import MagicMock
            import pytest

            class A:
                def __init__(self, name, value):
                    self.name = name
                    self.value = value

            def item_param():
                return 42

            circular_reference = A("circular_reference", A("child", None))
            circular_reference.value.value = circular_reference

            @pytest.mark.parametrize(
            'item',
            [
                pytest.param(A("test_name", "value"), marks=pytest.mark.skip),
                pytest.param(A("test_name", A("inner_name", "value")), marks=pytest.mark.skip),
                pytest.param(item_param, marks=pytest.mark.skip),
                pytest.param({"a": A("test_name", "value"), "b": [1, 2, 3]}, marks=pytest.mark.skip),
                pytest.param(MagicMock(value=MagicMock()), marks=pytest.mark.skip),
                pytest.param(circular_reference, marks=pytest.mark.skip),
                pytest.param({("x", "y"): 12345}, marks=pytest.mark.skip)
            ]
            )
            class Test1(object):
                def test_1(self, item):
                    assert item in {1, 2, 3}
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(skipped=7)
        spans = self.pop_spans()

        # Since object will have arbitrary addresses, only need to ensure that
        # the params string contains most of the string representation of the object.
        expected_params_contains = [
            "test_parameterize_case_complex_objects.A",
            "test_parameterize_case_complex_objects.A",
            "<function item_param>",
            "'a': <test_parameterize_case_complex_objects.A",
            "<MagicMock id=",
            "test_parameterize_case_complex_objects.A",
            "{('x', 'y'): 12345}",
        ]
        assert len(spans) == 7
        for i in range(len(expected_params_contains)):
            assert expected_params_contains[i] in spans[i].get_tag(test.PARAMETERS)

    def test_parameterize_case_encoding_error(self):
        """Test parametrize case with complex objects that cannot be JSON encoded."""
        py_file = self.testdir.makepyfile(
            """
            from mock import MagicMock
            import pytest

            class A:
                def __repr__(self):
                    raise Exception("Cannot __repr__")

            @pytest.mark.parametrize('item',[A()])
            class Test1(object):
                def test_1(self, item):
                    assert True
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert json.loads(spans[0].get_tag(test.PARAMETERS)) == {
            "arguments": {"item": "Could not encode"},
            "metadata": {},
        }

    def test_skip(self):
        """Test skip case."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.skip(reason="decorator")
            def test_decorator():
                pass

            def test_body():
                pytest.skip("body")
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(skipped=2)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert spans[0].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[0].get_tag(test.SKIP_REASON) == "decorator"
        assert spans[1].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[1].get_tag(test.SKIP_REASON) == "body"

    def test_skip_module_with_xfail_cases(self):
        """Test Xfail test cases for a module that is skipped entirely, which should be treated as skip tests."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            pytestmark = pytest.mark.skip(reason="reason")

            @pytest.mark.xfail(reason="XFail Case")
            def test_xfail():
                pass

            @pytest.mark.xfail(condition=False, reason="XFail Case")
            def test_xfail_conditional():
                pass
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(skipped=2)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert spans[0].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[0].get_tag(test.SKIP_REASON) == "reason"
        assert spans[1].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[1].get_tag(test.SKIP_REASON) == "reason"

    def test_skipif_module(self):
        """Test XFail test cases for a module that is skipped entirely with the skipif marker."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            pytestmark = pytest.mark.skipif(True, reason="reason")

            @pytest.mark.xfail(reason="XFail")
            def test_xfail():
                pass

            @pytest.mark.xfail(condition=False, reason="XFail Case")
            def test_xfail_conditional():
                pass
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(skipped=2)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert spans[0].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[0].get_tag(test.SKIP_REASON) == "reason"
        assert spans[1].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[1].get_tag(test.SKIP_REASON) == "reason"

    def test_xfail_fails(self):
        """Test xfail (expected failure) which fails, should be marked as pass."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason="test should fail")
            def test_should_fail():
                assert 0

            @pytest.mark.xfail(condition=True, reason="test should xfail")
            def test_xfail_conditional():
                assert 0
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        # pytest records xfail as skipped
        rec.assertoutcome(skipped=2)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value
        assert spans[0].get_tag(test.RESULT) == test.Status.XFAIL.value
        assert spans[0].get_tag(XFAIL_REASON) == "test should fail"
        assert spans[1].get_tag(test.STATUS) == test.Status.PASS.value
        assert spans[1].get_tag(test.RESULT) == test.Status.XFAIL.value
        assert spans[1].get_tag(XFAIL_REASON) == "test should xfail"

    def test_xfail_runxfail_fails(self):
        """Test xfail with --runxfail flags should not crash when failing."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason='should fail')
            def test_should_fail():
                assert 0

        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", "--runxfail", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag(test.STATUS) == test.Status.FAIL.value

    def test_xfail_runxfail_passes(self):
        """Test xfail with --runxfail flags should not crash when passing."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason='should fail')
            def test_should_pass():
                assert 1

        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", "--runxfail", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value

    def test_xpass_not_strict(self):
        """Test xpass (unexpected passing) with strict=False, should be marked as pass."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason="test should fail")
            def test_should_fail_but_passes():
                pass

            @pytest.mark.xfail(condition=True, reason="test should not xfail")
            def test_should_not_fail():
                pass
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=2)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value
        assert spans[0].get_tag(test.RESULT) == test.Status.XPASS.value
        assert spans[0].get_tag(XFAIL_REASON) == "test should fail"
        assert spans[1].get_tag(test.STATUS) == test.Status.PASS.value
        assert spans[1].get_tag(test.RESULT) == test.Status.XPASS.value
        assert spans[1].get_tag(XFAIL_REASON) == "test should not xfail"

    def test_xpass_strict(self):
        """Test xpass (unexpected passing) with strict=True, should be marked as fail."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason="test should fail", strict=True)
            def test_should_fail():
                pass
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(failed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag(test.STATUS) == test.Status.FAIL.value
        assert spans[0].get_tag(test.RESULT) == test.Status.XPASS.value
        # Note: XFail (strict=True) does not mark the reason with result.wasxfail but into result.longrepr,
        # however it provides the entire traceback/error into longrepr.
        assert "test should fail" in spans[0].get_tag(XFAIL_REASON)

    def test_tags(self):
        """Test ddspan tags."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.dd_tags(mark="dd_tags")
            def test_fixture(ddspan):
                assert ddspan is not None
                ddspan.set_tag("world", "hello")
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag("world") == "hello"
        assert spans[0].get_tag("mark") == "dd_tags"
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value

    def test_service_name_repository_name(self):
        """Test span's service name is set to repository name."""
        self.monkeypatch.setenv("APPVEYOR", "true")
        self.monkeypatch.setenv("APPVEYOR_REPO_PROVIDER", "github")
        self.monkeypatch.setenv("APPVEYOR_REPO_NAME", "test-repository-name")
        py_file = self.testdir.makepyfile(
            """
            import os

            def test_service(ddspan):
                assert 'test-repository-name' == ddspan.service
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.subprocess_run("--ddtrace", file_name)
        rec.assert_outcomes(passed=1)

    def test_default_service_name(self):
        """Test default service name if no repository name found."""
        providers = [provider for (provider, extract) in ci.PROVIDERS]
        for provider in providers:
            self.monkeypatch.delenv(provider, raising=False)
        py_file = self.testdir.makepyfile(
            """
            def test_service(ddspan):
                assert ddspan.service == "pytest"
                assert ddspan.name == "pytest.test"
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.subprocess_run("--ddtrace", file_name)
        rec.assert_outcomes(passed=1)

    def test_dd_service_name(self):
        """Test dd service name."""
        self.monkeypatch.setenv("DD_SERVICE", "mysvc")
        if "DD_PYTEST_SERVICE" in os.environ:
            self.monkeypatch.delenv("DD_PYTEST_SERVICE")

        py_file = self.testdir.makepyfile(
            """
            import os

            def test_service(ddspan):
                assert 'mysvc' == os.getenv('DD_SERVICE')
                assert os.getenv('DD_PYTEST_SERVICE') is None
                assert 'mysvc' == ddspan.service
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.subprocess_run("--ddtrace", file_name)
        assert 0 == rec.ret

    def test_dd_pytest_service_name(self):
        """Test integration service name."""
        self.monkeypatch.setenv("DD_SERVICE", "mysvc")
        self.monkeypatch.setenv("DD_PYTEST_SERVICE", "pymysvc")
        self.monkeypatch.setenv("DD_PYTEST_OPERATION_NAME", "mytest")

        py_file = self.testdir.makepyfile(
            """
            import os

            def test_service(ddspan):
                assert 'mysvc' == os.getenv('DD_SERVICE')
                assert 'pymysvc' == os.getenv('DD_PYTEST_SERVICE')
                assert 'pymysvc' == ddspan.service
                assert 'mytest' == ddspan.name
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.subprocess_run("--ddtrace", file_name)
        assert 0 == rec.ret

    def test_dd_origin_tag_propagated_to_every_span(self):
        """Test that every span in generated trace has the dd_origin tag."""
        py_file = self.testdir.makepyfile(
            """
            import pytest
            import ddtrace
            from ddtrace import Pin

            def test_service(ddspan, pytestconfig):
                tracer = Pin.get_from(pytestconfig).tracer
                with tracer.trace("SPAN2") as span2:
                    with tracer.trace("SPAN3") as span3:
                        with tracer.trace("SPAN4") as span4:
                            assert True
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)

        spans = self.pop_spans()
        # Check if spans tagged with dd_origin after encoding and decoding as the tagging occurs at encode time
        encoder = self.tracer.encoder
        encoder.put(spans)
        trace = encoder.encode()
        (decoded_trace,) = self.tracer.encoder._decode(trace)
        assert len(decoded_trace) == 4
        for span in decoded_trace:
            assert span[b"meta"][b"_dd.origin"] == b"ciapp-test"

    def test_pytest_doctest_module(self):
        """Test that pytest with doctest works as expected."""
        py_file = self.testdir.makepyfile(
            """
        '''
        This module supplies one function foo(). For example,
        >>> foo()
        42
        '''

        def foo():
            '''Returns the answer to life, the universe, and everything.
            >>> foo()
            42
            '''
            return 42

        def test_foo():
            assert foo() == 42
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", "--doctest-modules", file_name)
        rec.assertoutcome(passed=3)
        spans = self.pop_spans()

        assert len(spans) == 3
        for span in spans:
            assert span.get_tag(test.SUITE) == file_name.partition(".py")[0]

    def test_pytest_sets_sample_priority(self):
        """Test sample priority tags."""
        py_file = self.testdir.makepyfile(
            """
            def test_sample_priority():
                assert True is True
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_metric(SAMPLING_PRIORITY_KEY) == 1

    def test_pytest_exception(self):
        """Test that pytest sets exception information correctly."""
        py_file = self.testdir.makepyfile(
            """
        def test_will_fail():
            assert 2 == 1
        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        test_span = spans[0]
        assert test_span.get_tag(test.STATUS) == test.Status.FAIL.value
        assert test_span.get_tag("error.type").endswith("AssertionError") is True
        assert test_span.get_tag("error.msg") == "assert 2 == 1"
        assert test_span.get_tag("error.stack") is not None

    def test_pytest_tests_with_internal_exceptions_get_test_status(self):
        """Test that pytest sets a fail test status if it has an internal exception."""
        py_file = self.testdir.makepyfile(
            """
        import pytest

        # This is bad usage and results in a pytest internal exception
        @pytest.mark.filterwarnings("ignore::pytest.ExceptionThatDoesNotExist")
        def test_will_fail_internally():
            assert 2 == 2
        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        test_span = spans[0]
        assert test_span.get_tag(test.STATUS) == test.Status.FAIL.value
        assert test_span.get_tag("error.type") is None

    def test_pytest_broken_setup_will_be_reported_as_error(self):
        """Test that pytest sets a fail test status if the setup fails."""
        py_file = self.testdir.makepyfile(
            """
        import pytest

        @pytest.fixture
        def my_fixture():
            raise Exception('will fail in setup')
            yield

        def test_will_fail_in_setup(my_fixture):
            assert 1 == 1
        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        test_span = spans[0]

        assert test_span.get_tag(test.STATUS) == test.Status.FAIL.value
        assert test_span.get_tag("error.type").endswith("Exception") is True
        assert test_span.get_tag("error.msg") == "will fail in setup"
        assert test_span.get_tag("error.stack") is not None

    def test_pytest_broken_teardown_will_be_reported_as_error(self):
        """Test that pytest sets a fail test status if the teardown fails."""
        py_file = self.testdir.makepyfile(
            """
        import pytest

        @pytest.fixture
        def my_fixture():
            yield
            raise Exception('will fail in teardown')

        def test_will_fail_in_teardown(my_fixture):
            assert 1 == 1
        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        test_span = spans[0]

        assert test_span.get_tag(test.STATUS) == test.Status.FAIL.value
        assert test_span.get_tag("error.type").endswith("Exception") is True
        assert test_span.get_tag("error.msg") == "will fail in teardown"
        assert test_span.get_tag("error.stack") is not None

    def test_pytest_will_report_its_version(self):
        py_file = self.testdir.makepyfile(
            """
        import pytest

        def test_will_work():
            assert 1 == 1
        """
        )
        file_name = os.path.basename(py_file.strpath)
        self.inline_run("--ddtrace", file_name)
        spans = self.pop_spans()

        assert len(spans) == 1
        test_span = spans[0]

        assert test_span.get_tag(test.FRAMEWORK_VERSION) == pytest.__version__

    def test_pytest_will_report_codeowners(self):
        file_names = []
        py_team_a_file = self.testdir.makepyfile(
            test_team_a="""
        import pytest

        def test_team_a():
            assert 1 == 1
        """
        )
        file_names.append(os.path.basename(py_team_a_file.strpath))
        py_team_b_file = self.testdir.makepyfile(
            test_team_b="""
        import pytest

        def test_team_b():
            assert 1 == 1
        """
        )
        file_names.append(os.path.basename(py_team_b_file.strpath))
        codeowners = "* @default-team\n{0} @team-b @backup-b".format(os.path.basename(py_team_b_file.strpath))
        self.testdir.makefile("", CODEOWNERS=codeowners)

        self.inline_run("--ddtrace", *file_names)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert json.loads(spans[0].get_tag(test.CODEOWNERS)) == ["@default-team"], spans[0]
        assert json.loads(spans[1].get_tag(test.CODEOWNERS)) == ["@team-b", "@backup-b"], spans[1]


@pytest.mark.parametrize(
    "repository_url,repository_name",
    [
        ("https://github.com/DataDog/dd-trace-py.git", "dd-trace-py"),
        ("https://github.com/DataDog/dd-trace-py", "dd-trace-py"),
        ("git@github.com:DataDog/dd-trace-py.git", "dd-trace-py"),
        ("git@github.com:DataDog/dd-trace-py", "dd-trace-py"),
        ("dd-trace-py", "dd-trace-py"),
        ("git@hostname.com:org/repo-name.git", "repo-name"),
        ("git@hostname.com:org/repo-name", "repo-name"),
        ("ssh://git@hostname.com:org/repo-name", "repo-name"),
        ("git+git://github.com/org/repo-name.git", "repo-name"),
        ("git+ssh://github.com/org/repo-name.git", "repo-name"),
        ("git+https://github.com/org/repo-name.git", "repo-name"),
    ],
)
def test_repository_name_extracted(repository_url, repository_name):
    assert _extract_repository_name(repository_url) == repository_name


def test_repository_name_not_extracted_warning():
    """If provided an invalid repository url, should raise warning and return original repository url"""
    repository_url = "https://github.com:organ[ization/repository-name"
    with mock.patch("ddtrace.contrib.pytest.plugin.log") as mock_log:
        extracted_repository_name = _extract_repository_name(repository_url)
        assert extracted_repository_name == repository_url
    mock_log.warning.assert_called_once_with("Repository name cannot be parsed from repository_url: %s", repository_url)



__all__ = ["get_extensions"]

import contextlib
import io
import os
import platform
from setuptools import Extension
import shutil
import sys
import tempfile

POSIX = os.name == "posix"
WINDOWS = os.name == "nt"
LINUX = sys.platform.startswith("linux")
MACOS = sys.platform.startswith("darwin")
OSX = MACOS  # deprecated alias
FREEBSD = sys.platform.startswith("freebsd")
OPENBSD = sys.platform.startswith("openbsd")
NETBSD = sys.platform.startswith("netbsd")
BSD = FREEBSD or OPENBSD or NETBSD
SUNOS = sys.platform.startswith(("sunos", "solaris"))
AIX = sys.platform.startswith("aix")


@contextlib.contextmanager
def silenced_output(stream_name):
    class DummyFile(io.BytesIO):
        # see: https://github.com/giampaolo/psutil/issues/678
        errors = "ignore"

        def write(self, s):
            pass

    orig = getattr(sys, stream_name)
    try:
        setattr(sys, stream_name, DummyFile())
        yield
    finally:
        setattr(sys, stream_name, orig)


def get_extensions():
    macros = [("PSUTIL_VERSION", 567)]
    if POSIX:
        macros.append(("PSUTIL_POSIX", 1))
    if BSD:
        macros.append(("PSUTIL_BSD", 1))

    sources = ["ddtrace/vendor/psutil/_psutil_common.c"]
    if POSIX:
        sources.append("ddtrace/vendor/psutil/_psutil_posix.c")

    if WINDOWS:

        def get_winver():
            win_maj, win_min = sys.getwindowsversion()[0:2]
            return "0x0%s" % ((win_maj * 100) + win_min)

        if sys.getwindowsversion()[0] < 6:
            msg = "this Windows version is too old (< Windows Vista); "
            msg += "psutil 3.4.2 is the latest version which supports Windows "
            msg += "2000, XP and 2003 server"
            raise RuntimeError(msg)

        macros.append(("PSUTIL_WINDOWS", 1))
        macros.extend(
            [
                # be nice to mingw, see:
                # http://www.mingw.org/wiki/Use_more_recent_defined_functions
                ("_WIN32_WINNT", get_winver()),
                ("_AVAIL_WINVER_", get_winver()),
                ("_CRT_SECURE_NO_WARNINGS", None),
                # see: https://github.com/giampaolo/psutil/issues/348
                ("PSAPI_VERSION", 1),
            ]
        )

        sources += [
            "ddtrace/vendor/psutil/_psutil_windows.c",
            "ddtrace/vendor/psutil/arch/windows/process_info.c",
            "ddtrace/vendor/psutil/arch/windows/process_handles.c",
            "ddtrace/vendor/psutil/arch/windows/security.c",
            "ddtrace/vendor/psutil/arch/windows/inet_ntop.c",
            "ddtrace/vendor/psutil/arch/windows/services.c",
            "ddtrace/vendor/psutil/arch/windows/global.c",
            "ddtrace/vendor/psutil/arch/windows/wmi.c",
        ]
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_windows",
            sources=sources,
            define_macros=macros,
            libraries=[
                "psapi",
                "kernel32",
                "advapi32",
                "shell32",
                "netapi32",
                "wtsapi32",
                "ws2_32",
                "PowrProf",
                "pdh",
            ],
            # extra_compile_args=["/Z7"],
            # extra_link_args=["/DEBUG"]
        )

    elif MACOS:
        macros.append(("PSUTIL_OSX", 1))
        sources += [
            "ddtrace/vendor/psutil/_psutil_osx.c",
            "ddtrace/vendor/psutil/arch/osx/process_info.c",
        ]
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_osx",
            sources=sources,
            define_macros=macros,
            extra_link_args=["-framework", "CoreFoundation", "-framework", "IOKit"],
        )

    elif FREEBSD:
        macros.append(("PSUTIL_FREEBSD", 1))
        sources += [
            "ddtrace/vendor/psutil/_psutil_bsd.c",
            "ddtrace/vendor/psutil/arch/freebsd/specific.c",
            "ddtrace/vendor/psutil/arch/freebsd/sys_socks.c",
            "ddtrace/vendor/psutil/arch/freebsd/proc_socks.c",
        ]
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_bsd", sources=sources, define_macros=macros, libraries=["devstat"],
        )

    elif OPENBSD:
        macros.append(("PSUTIL_OPENBSD", 1))
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_bsd",
            sources=sources + ["ddtrace/vendor/psutil/_psutil_bsd.c", "ddtrace/vendor/psutil/arch/openbsd/specific.c"],
            define_macros=macros,
            libraries=["kvm"],
        )

    elif NETBSD:
        macros.append(("PSUTIL_NETBSD", 1))
        sources += [
            "ddtrace/vendor/psutil/_psutil_bsd.c",
            "ddtrace/vendor/psutil/arch/netbsd/specific.c",
            "ddtrace/vendor/psutil/arch/netbsd/socks.c",
        ]
        ext = Extension("ddtrace.vendor.psutil._psutil_bsd", sources=sources, define_macros=macros, libraries=["kvm"],)

    elif LINUX:

        def get_ethtool_macro():
            # see: https://github.com/giampaolo/ddtrace/vendor/psutil/issues/659
            from distutils.unixccompiler import UnixCCompiler
            from distutils.errors import CompileError

            with tempfile.NamedTemporaryFile(suffix=".c", delete=False, mode="wt") as f:
                f.write("#include <linux/ethtool.h>")

            output_dir = tempfile.mkdtemp()
            try:
                compiler = UnixCCompiler()
                # https://github.com/giampaolo/ddtrace/vendor/psutil/pull/1568
                if os.getenv("CC"):
                    compiler.set_executable("compiler_so", os.getenv("CC"))
                with silenced_output("stderr"):
                    with silenced_output("stdout"):
                        compiler.compile([f.name], output_dir=output_dir)
            except CompileError:
                return ("PSUTIL_ETHTOOL_MISSING_TYPES", 1)
            else:
                return None
            finally:
                os.remove(f.name)
                shutil.rmtree(output_dir)

        macros.append(("PSUTIL_LINUX", 1))
        ETHTOOL_MACRO = get_ethtool_macro()
        if ETHTOOL_MACRO is not None:
            macros.append(ETHTOOL_MACRO)
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_linux",
            sources=sources + ["ddtrace/vendor/psutil/_psutil_linux.c"],
            define_macros=macros,
        )

    elif SUNOS:
        macros.append(("PSUTIL_SUNOS", 1))
        sources += [
            "ddtrace/vendor/psutil/_psutil_sunos.c",
            "ddtrace/vendor/psutil/arch/solaris/v10/ifaddrs.c",
            "ddtrace/vendor/psutil/arch/solaris/environ.c",
        ]
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_sunos",
            sources=sources,
            define_macros=macros,
            libraries=["kstat", "nsl", "socket"],
        )

    elif AIX:
        macros.append(("PSUTIL_AIX", 1))
        sources += [
            "ddtrace/vendor/psutil/_psutil_aix.c",
            "ddtrace/vendor/psutil/arch/aix/net_connections.c",
            "ddtrace/vendor/psutil/arch/aix/common.c",
            "ddtrace/vendor/psutil/arch/aix/ifaddrs.c",
        ]
        ext = Extension(
            "ddtrace.vendor.psutil._psutil_aix", sources=sources, libraries=["perfstat"], define_macros=macros,
        )
    else:
        raise RuntimeError("platform %s is not supported" % sys.platform)

    if POSIX:
        posix_extension = Extension("ddtrace.vendor.psutil._psutil_posix", define_macros=macros, sources=sources)
        if SUNOS:
            posix_extension.libraries.append("socket")
            if platform.release() == "5.10":
                posix_extension.sources.append("ddtrace/vendor/psutil/arch/solaris/v10/ifaddrs.c")
                posix_extension.define_macros.append(("PSUTIL_SUNOS10", 1))
        elif AIX:
            posix_extension.sources.append("ddtrace/vendor/psutil/arch/aix/ifaddrs.c")

        return [ext, posix_extension]
    else:
        return [ext]



# -*- encoding: utf-8 -*-
"""
An API to provide atexit functionalities
"""
from __future__ import absolute_import

import atexit
import logging
import typing


log = logging.getLogger(__name__)


if hasattr(atexit, "unregister"):
    register = atexit.register
    unregister = atexit.unregister
else:
    # Hello Python 2!
    _registry = []  # type: typing.List[typing.Tuple[typing.Callable[..., None], typing.Tuple, typing.Dict]]

    def _ddtrace_atexit():
        # type: (...) -> None
        """Wrapper function that calls all registered function on normal program termination"""
        global _registry

        # DEV: we make a copy of the registry to prevent hook execution from
        # introducing new hooks, potentially causing an infinite loop.
        for hook, args, kwargs in list(_registry):
            try:
                hook(*args, **kwargs)
            except Exception:
                # Mimic the behaviour of Python's atexit hooks.
                log.exception("Error in atexit hook %r", hook)

    def register(
        func,  # type: typing.Callable[..., typing.Any]
        *args,  # type: typing.Any
        **kwargs  # type: typing.Any
    ):
        # type: (...) -> typing.Callable[..., typing.Any]
        """Register a function to be executed upon normal program termination"""

        _registry.append((func, args, kwargs))
        return func

    def unregister(func):
        # type: (typing.Callable[..., typing.Any]) -> None
        """
        Unregister an exit function which was previously registered using
        atexit.register.
        If the function was not registered, it is ignored.
        """
        global _registry

        _registry = [(f, args, kwargs) for f, args, kwargs in _registry if f != func]

    atexit.register(_ddtrace_atexit)



import sys
from types import FunctionType
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import cast

from six import PY3


try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore[misc]

from bytecode import Bytecode
from bytecode import Compare
from bytecode import CompilerFlags
from bytecode import Instr
from bytecode import Label

from .compat import PYTHON_VERSION_INFO as PY


class WrappedFunction(Protocol):
    """A wrapped function."""

    __dd_wrapped__ = None  # type: Optional[FunctionType]
    __dd_wrappers__ = None  # type: Optional[Dict[Any, Any]]

    def __call__(self, *args, **kwargs):
        pass


Wrapper = Callable[[FunctionType, Tuple[Any], Dict[str, Any]], Any]


def wrap_bytecode(wrapper, wrapped):
    # type: (Wrapper, FunctionType) -> Bytecode
    """Wrap a function with a wrapper function.

    The wrapper function expects the wrapped function as the first argument,
    followed by the tuple of arguments and the dictionary of keyword arguments.
    The nature of the wrapped function is also honored, meaning that a generator
    function will return a generator function, and a coroutine function will
    return a coroutine function, and so on. The signature is also preserved to
    avoid breaking, e.g., usages of the ``inspect`` module.
    """

    def compare_exc(label, lineno):
        """Compat helper for comparing exceptions."""
        return (
            Instr("COMPARE_OP", Compare.EXC_MATCH, lineno=lineno)
            if PY < (3, 9)
            else Instr("JUMP_IF_NOT_EXC_MATCH", label, lineno=lineno)
        )

    def jump_if_false(label, lineno):
        """Compat helper for jumping if false after comparing exceptions."""
        return Instr("POP_JUMP_IF_FALSE", label, lineno=lineno) if PY < (3, 9) else Instr("NOP", lineno=lineno)

    def end_finally(lineno):
        """Compat helper for ending finally blocks."""
        if PY < (3, 9):
            return Instr("END_FINALLY", lineno=lineno)
        elif PY < (3, 10):
            return Instr("RERAISE", lineno=lineno)
        return Instr("RERAISE", 0, lineno=lineno)

    code = wrapped.__code__
    lineno = code.co_firstlineno
    varargs = bool(code.co_flags & CompilerFlags.VARARGS)
    varkwargs = bool(code.co_flags & CompilerFlags.VARKEYWORDS)
    nargs = code.co_argcount
    argnames = code.co_varnames[:nargs]
    try:
        kwonlyargs = code.co_kwonlyargcount
    except AttributeError:
        kwonlyargs = 0
    kwonlyargnames = code.co_varnames[nargs : nargs + kwonlyargs]
    varargsname = code.co_varnames[nargs + kwonlyargs] if varargs else None
    varkwargsname = code.co_varnames[nargs + kwonlyargs + varargs] if varkwargs else None

    # Push the wrapper function that is to be called and the wrapped function to
    # be passed as first argument.
    instrs = [
        Instr("LOAD_CONST", wrapper, lineno=lineno),
        Instr("LOAD_CONST", wrapped, lineno=lineno),
    ]

    # Build the tuple of all the positional arguments
    if nargs:
        instrs.extend([Instr("LOAD_FAST", argname, lineno=lineno) for argname in argnames])
        instrs.append(Instr("BUILD_TUPLE", nargs, lineno=lineno))
        if varargs:
            instrs.extend(
                [
                    Instr("LOAD_FAST", varargsname, lineno=lineno),
                    Instr("INPLACE_ADD", lineno=lineno),
                ]
            )
    elif varargs:
        instrs.append(Instr("LOAD_FAST", varargsname, lineno=lineno))
    else:
        instrs.append(Instr("BUILD_TUPLE", 0, lineno=lineno))

    # Prepare the keyword arguments
    if kwonlyargs:
        for arg in kwonlyargnames:
            instrs.extend(
                [
                    Instr("LOAD_CONST", arg, lineno=lineno),
                    Instr("LOAD_FAST", arg, lineno=lineno),
                ]
            )
        instrs.append(Instr("BUILD_MAP", kwonlyargs, lineno=lineno))
        if varkwargs:
            instrs.extend(
                [
                    Instr("DUP_TOP", lineno=lineno),
                    Instr("LOAD_ATTR", "update", lineno=lineno),
                    Instr("LOAD_FAST", varkwargsname, lineno=lineno),
                    Instr("CALL_FUNCTION", 1, lineno=lineno),
                    Instr("POP_TOP", lineno=lineno),
                ]
            )

    elif varkwargs:
        instrs.append(Instr("LOAD_FAST", varkwargsname, lineno=lineno))

    else:
        instrs.append(Instr("BUILD_MAP", 0, lineno=lineno))

    # Call the wrapper function with the wrapped function, the positional and
    # keyword arguments, and return the result.
    instrs.extend(
        [
            Instr("CALL_FUNCTION", 3, lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
        ]
    )

    # If the function has special flags set, like the generator, async generator
    # or coroutine, inject unraveling code before the return opcode.
    if CompilerFlags.GENERATOR & code.co_flags and not (CompilerFlags.COROUTINE & code.co_flags):
        stopiter = Label()
        loop = Label()
        genexit = Label()
        exc = Label()
        propagate = Label()

        # DEV: This is roughly equivalent to
        #
        # __ddgen = wrapper(wrapped, args, kwargs)
        # __ddgensend = __ddgen.send
        # try:
        #     value = next(__ddgen)
        # except StopIteration:
        #     return
        # while True:
        #     try:
        #         tosend = yield value
        #     except GeneratorExit:
        #         return __ddgen.close()
        #     except:
        #         return __ddgen.throw(*sys.exc_info())
        #     try:
        #         value = __ddgensend(tosend)
        #     except StopIteration:
        #         return
        #
        instrs[-1:-1] = [
            Instr("DUP_TOP", lineno=lineno),
            Instr("STORE_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "send", lineno=lineno),
            Instr("STORE_FAST", "__ddgensend", lineno=lineno),
            Instr("LOAD_CONST", next, lineno=lineno),
            Instr("LOAD_FAST", "__ddgen", lineno=lineno),
            loop,
            Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", stopiter, lineno=lineno),
            Instr("CALL_FUNCTION", 1, lineno=lineno),
            Instr("POP_BLOCK", lineno=lineno),
            Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", genexit, lineno=lineno),
            Instr("YIELD_VALUE", lineno=lineno),
            Instr("POP_BLOCK", lineno=lineno),
            Instr("LOAD_FAST", "__ddgensend", lineno=lineno),
            Instr("ROT_TWO", lineno=lineno),
            Instr("JUMP_ABSOLUTE", loop, lineno=lineno),
            stopiter,  # except StpIteration:
            Instr("DUP_TOP", lineno=lineno),
            Instr("LOAD_CONST", StopIteration, lineno=lineno),
            compare_exc(propagate, lineno),
            jump_if_false(propagate, lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
            propagate,
            end_finally(lineno),
            Instr("LOAD_CONST", None, lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
            genexit,  # except GeneratorExit:
            Instr("DUP_TOP", lineno=lineno),
            Instr("LOAD_CONST", GeneratorExit, lineno=lineno),
            compare_exc(exc, lineno),
            jump_if_false(exc, lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "close", lineno=lineno),
            Instr("CALL_FUNCTION", 0, lineno=lineno),
            Instr("RETURN_VALUE", lineno=lineno),
            exc,  # except:
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("POP_TOP", lineno=lineno),
            Instr("LOAD_FAST", "__ddgen", lineno=lineno),
            Instr("LOAD_ATTR", "throw", lineno=lineno),
            Instr("LOAD_CONST", sys.exc_info, lineno=lineno),
            Instr("CALL_FUNCTION", 0, lineno=lineno),
            Instr("CALL_FUNCTION_VAR" if PY < (3, 6) else "CALL_FUNCTION_EX", 0, lineno=lineno),
        ]
    elif PY3:
        if CompilerFlags.COROUTINE & code.co_flags:
            # DEV: This is just
            # >>> return await wrapper(wrapped, args, kwargs)
            instrs[-1:-1] = [
                Instr("GET_AWAITABLE", lineno=lineno),
                Instr("LOAD_CONST", None, lineno=lineno),
                Instr("YIELD_FROM", lineno=lineno),
            ]
        elif CompilerFlags.ASYNC_GENERATOR & code.co_flags:
            stopiter = Label()
            loop = Label()
            genexit = Label()
            exc = Label()
            propagate = Label()

            # DEV: This is roughly equivalent to
            #
            # __ddgen = wrapper(wrapped, args, kwargs)
            # __ddgensend = __ddgen.asend
            # try:
            #     value = await _ddgen.__anext__()
            # except StopAsyncIteration:
            #     return
            # while True:
            #     try:
            #         tosend = yield value
            #     except GeneratorExit:
            #         __ddgen.close()
            #     except:
            #         __ddgen.throw(*sys.exc_info())
            #     try:
            #         value = await __ddgensend(tosend)
            #     except StopAsyncIteration:
            #         return
            #
            instrs[-1:-1] = [
                Instr("DUP_TOP", lineno=lineno),
                Instr("STORE_FAST", "__ddgen", lineno=lineno),
                Instr("LOAD_ATTR", "asend", lineno=lineno),
                Instr("STORE_FAST", "__ddgensend", lineno=lineno),
                Instr("LOAD_FAST", "__ddgen", lineno=lineno),
                Instr("LOAD_ATTR", "__anext__", lineno=lineno),
                Instr("CALL_FUNCTION", 0, lineno=lineno),
                loop,
                Instr("GET_AWAITABLE", lineno=lineno),
                Instr("LOAD_CONST", None, lineno=lineno),
                Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", stopiter, lineno=lineno),
                Instr("YIELD_FROM", lineno=lineno),
                Instr("POP_BLOCK", lineno=lineno),
                Instr("SETUP_EXCEPT" if PY < (3, 8) else "SETUP_FINALLY", genexit, lineno=lineno),
                Instr("YIELD_VALUE", lineno=lineno),
                Instr("POP_BLOCK", lineno=lineno),
                Instr("LOAD_FAST", "__ddgensend", lineno=lineno),
                Instr("ROT_TWO", lineno=lineno),
                Instr("CALL_FUNCTION", 1, lineno=lineno),
                Instr("JUMP_ABSOLUTE", loop, lineno=lineno),
                stopiter,  # except StopAsyncIteration:
                Instr("DUP_TOP", lineno=lineno),
                Instr("LOAD_CONST", StopAsyncIteration, lineno=lineno),
                compare_exc(propagate, lineno),
                jump_if_false(propagate, lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("LOAD_CONST", None, lineno=lineno),
                Instr("RETURN_VALUE", lineno=lineno),
                propagate,  # finally:
                end_finally(lineno),
                Instr("LOAD_CONST", None, lineno=lineno),
                Instr("RETURN_VALUE", lineno=lineno),
                genexit,  # except GeneratorExit:
                Instr("DUP_TOP", lineno=lineno),
                Instr("LOAD_CONST", GeneratorExit, lineno=lineno),
                compare_exc(exc, lineno),
                jump_if_false(exc, lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("LOAD_FAST", "__ddgen", lineno=lineno),
                Instr("LOAD_ATTR", "aclose", lineno=lineno),
                Instr("CALL_FUNCTION", 0, lineno=lineno),
                Instr("GET_AWAITABLE", lineno=lineno),
                Instr("LOAD_CONST", None, lineno=lineno),
                Instr("YIELD_FROM", lineno=lineno),
                Instr("POP_EXCEPT", lineno=lineno),
                Instr("RETURN_VALUE", lineno=lineno),
                exc,  # except:
                Instr("POP_TOP", lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("POP_TOP", lineno=lineno),
                Instr("LOAD_FAST", "__ddgen", lineno=lineno),
                Instr("LOAD_ATTR", "athrow", lineno=lineno),
                Instr("LOAD_CONST", sys.exc_info, lineno=lineno),
                Instr("CALL_FUNCTION", 0, lineno=lineno),
                Instr("CALL_FUNCTION_EX", 0, lineno=lineno),
                Instr("GET_AWAITABLE", lineno=lineno),
                Instr("LOAD_CONST", None, lineno=lineno),
                Instr("YIELD_FROM", lineno=lineno),
                Instr("POP_EXCEPT", lineno=lineno),
            ]

    return Bytecode(instrs)


def wrap(f, wrapper):
    # type: (FunctionType, Wrapper) -> WrappedFunction
    """Wrap a function with a wrapper.

    The wrapper expects the function as first argument, followed by the tuple
    of positional arguments and the dict of keyword arguments.

    Note that this changes the behavior of the original function with the
    wrapper function, instead of creating a new function object.
    """
    wrapped = FunctionType(
        f.__code__,
        f.__globals__,
        "<wrapped>",
        f.__defaults__,
        f.__closure__,
    )
    try:
        wf = cast(WrappedFunction, f)
        cast(WrappedFunction, wrapped).__dd_wrapped__ = cast(FunctionType, wf.__dd_wrapped__)
    except AttributeError:
        pass

    if PY3:
        wrapped.__kwdefaults__ = f.__kwdefaults__

    code = wrap_bytecode(wrapper, wrapped)
    code.freevars = f.__code__.co_freevars
    code.name = f.__code__.co_name
    code.filename = f.__code__.co_filename
    code.flags = f.__code__.co_flags
    code.argcount = f.__code__.co_argcount
    try:
        code.posonlyargcount = f.__code__.co_posonlyargcount
    except AttributeError:
        pass

    nargs = code.argcount
    try:
        code.kwonlyargcount = f.__code__.co_kwonlyargcount
        nargs += code.kwonlyargcount
    except AttributeError:
        pass
    nargs += bool(code.flags & CompilerFlags.VARARGS) + bool(code.flags & CompilerFlags.VARKEYWORDS)
    code.argnames = f.__code__.co_varnames[:nargs]

    f.__code__ = code.to_code()

    # DEV: Multiple wrapping is implemented as a singly-linked list via the
    # __dd_wrapped__ attribute.
    wf = cast(WrappedFunction, f)
    wf.__dd_wrapped__ = wrapped

    return wf


def unwrap(wf, wrapper):
    # type: (WrappedFunction, Wrapper) -> FunctionType
    """Unwrap a wrapped function.

    This is the reverse of :func:`wrap`. In case of multiple wrapping layers,
    this will unwrap the one that uses ``wrapper``. If the function was not
    wrapped with ``wrapper``, it will return the first argument.
    """
    # DEV: Multiple wrapping layers are singly-linked via __dd_wrapped__. When
    # we find the layer that needs to be removed we also have to ensure that we
    # update the link at the deletion site if there is a non-empty tail.
    try:
        inner = cast(FunctionType, wf.__dd_wrapped__)

        # Sanity check
        assert inner.__name__ == "<wrapped>", "Wrapper has wrapped function"

        if wrapper not in cast(FunctionType, wf).__code__.co_consts:
            # This is not the correct wrapping layer. Try with the next one.
            inner_wf = cast(WrappedFunction, inner)
            return unwrap(inner_wf, wrapper)

        # Remove the current wrapping layer by moving the next one over the
        # current one.
        f = cast(FunctionType, wf)
        f.__code__ = inner.__code__
        try:
            # Update the link to the next layer.
            inner_wf = cast(WrappedFunction, inner)
            wf.__dd_wrapped__ = inner_wf.__dd_wrapped__  # type: ignore[assignment]
        except AttributeError:
            # No more wrapping layers. Restore the original function by removing
            # this extra attribute.
            del wf.__dd_wrapped__

        return f

    except AttributeError:
        # The function is not wrapped so we return it as is.
        return cast(FunctionType, wf)



import pymysql

from ddtrace import Pin
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.contrib.pymysql.patch import patch
from ddtrace.contrib.pymysql.patch import unpatch
from ddtrace.internal.compat import PY2
from ddtrace.internal.compat import stringify
from tests.opentracer.utils import init_tracer
from tests.utils import TracerTestCase
from tests.utils import assert_dict_issuperset
from tests.utils import assert_is_measured

from ...contrib.config import MYSQL_CONFIG


class PyMySQLCore(object):
    """PyMySQL test case reuses the connection across tests"""

    conn = None

    DB_INFO = {
        "out.host": MYSQL_CONFIG.get("host"),
    }
    if PY2:
        DB_INFO.update({"db.user": MYSQL_CONFIG.get("user"), "db.name": MYSQL_CONFIG.get("database")})
    else:
        DB_INFO.update(
            {
                "db.user": stringify(bytes(MYSQL_CONFIG.get("user"), encoding="utf-8")),
                "db.name": stringify(bytes(MYSQL_CONFIG.get("database"), encoding="utf-8")),
            }
        )

    def setUp(self):
        super(PyMySQLCore, self).setUp()
        patch()

    def tearDown(self):
        super(PyMySQLCore, self).tearDown()
        if self.conn and not self.conn._closed:
            self.conn.close()
        unpatch()

    def _get_conn_tracer(self):
        # implement me
        pass

    def test_simple_query(self):
        conn, tracer = self._get_conn_tracer()

        cursor = conn.cursor()

        # PyMySQL returns back the rowcount instead of a cursor
        rowcount = cursor.execute("SELECT 1")
        assert rowcount == 1

        rows = cursor.fetchall()
        assert len(rows) == 1
        spans = tracer.pop()
        assert len(spans) == 1

        span = spans[0]
        assert_is_measured(span)
        assert span.service == "pymysql"
        assert span.name == "pymysql.query"
        assert span.span_type == "sql"
        assert span.error == 0
        assert span.get_metric("out.port") == MYSQL_CONFIG.get("port")
        meta = {}
        meta.update(self.DB_INFO)
        assert_dict_issuperset(span.get_tags(), meta)

    def test_simple_query_fetchall(self):
        with self.override_config("pymysql", dict(trace_fetch_methods=True)):
            conn, tracer = self._get_conn_tracer()

            cursor = conn.cursor()
            cursor.execute("SELECT 1")
            rows = cursor.fetchall()
            assert len(rows) == 1
            spans = tracer.pop()
            assert len(spans) == 2

            span = spans[0]
            assert_is_measured(span)
            assert span.service == "pymysql"
            assert span.name == "pymysql.query"
            assert span.span_type == "sql"
            assert span.error == 0
            assert span.get_metric("out.port") == MYSQL_CONFIG.get("port")
            meta = {}
            meta.update(self.DB_INFO)
            assert_dict_issuperset(span.get_tags(), meta)

            fetch_span = spans[1]
            assert fetch_span.name == "pymysql.query.fetchall"

    def test_query_with_several_rows(self):
        conn, tracer = self._get_conn_tracer()

        cursor = conn.cursor()
        query = "SELECT n FROM (SELECT 42 n UNION SELECT 421 UNION SELECT 4210) m"
        cursor.execute(query)
        rows = cursor.fetchall()
        assert len(rows) == 3
        spans = tracer.pop()
        assert len(spans) == 1
        self.assertEqual(spans[0].name, "pymysql.query")

    def test_query_with_several_rows_fetchall(self):
        with self.override_config("pymysql", dict(trace_fetch_methods=True)):
            conn, tracer = self._get_conn_tracer()

            cursor = conn.cursor()
            query = "SELECT n FROM (SELECT 42 n UNION SELECT 421 UNION SELECT 4210) m"
            cursor.execute(query)
            rows = cursor.fetchall()
            assert len(rows) == 3
            spans = tracer.pop()
            assert len(spans) == 2

            fetch_span = spans[1]
            assert fetch_span.name == "pymysql.query.fetchall"

    def test_query_many(self):
        # tests that the executemany method is correctly wrapped.
        conn, tracer = self._get_conn_tracer()

        tracer.enabled = False
        cursor = conn.cursor()

        cursor.execute(
            """
            create table if not exists dummy (
                dummy_key VARCHAR(32) PRIMARY KEY,
                dummy_value TEXT NOT NULL)"""
        )
        tracer.enabled = True

        stmt = "INSERT INTO dummy (dummy_key, dummy_value) VALUES (%s, %s)"
        data = [("foo", "this is foo"), ("bar", "this is bar")]

        # PyMySQL `executemany()` returns the rowcount
        rowcount = cursor.executemany(stmt, data)
        assert rowcount == 2

        query = "SELECT dummy_key, dummy_value FROM dummy ORDER BY dummy_key"
        cursor.execute(query)
        rows = cursor.fetchall()
        assert len(rows) == 2
        assert rows[0][0] == "bar"
        assert rows[0][1] == "this is bar"
        assert rows[1][0] == "foo"
        assert rows[1][1] == "this is foo"

        spans = tracer.pop()
        assert len(spans) == 2
        cursor.execute("drop table if exists dummy")

    def test_query_many_fetchall(self):
        with self.override_config("pymysql", dict(trace_fetch_methods=True)):
            # tests that the executemany method is correctly wrapped.
            conn, tracer = self._get_conn_tracer()

            tracer.enabled = False
            cursor = conn.cursor()

            cursor.execute(
                """
                create table if not exists dummy (
                    dummy_key VARCHAR(32) PRIMARY KEY,
                    dummy_value TEXT NOT NULL)"""
            )
            tracer.enabled = True

            stmt = "INSERT INTO dummy (dummy_key, dummy_value) VALUES (%s, %s)"
            data = [("foo", "this is foo"), ("bar", "this is bar")]
            cursor.executemany(stmt, data)
            query = "SELECT dummy_key, dummy_value FROM dummy ORDER BY dummy_key"
            cursor.execute(query)
            rows = cursor.fetchall()
            assert len(rows) == 2
            assert rows[0][0] == "bar"
            assert rows[0][1] == "this is bar"
            assert rows[1][0] == "foo"
            assert rows[1][1] == "this is foo"

            spans = tracer.pop()
            assert len(spans) == 3
            cursor.execute("drop table if exists dummy")

            fetch_span = spans[2]
            assert fetch_span.name == "pymysql.query.fetchall"

    def test_query_proc(self):
        conn, tracer = self._get_conn_tracer()

        # create a procedure
        tracer.enabled = False
        cursor = conn.cursor()
        cursor.execute("DROP PROCEDURE IF EXISTS sp_sum")
        cursor.execute(
            """
            CREATE PROCEDURE sp_sum (IN p1 INTEGER, IN p2 INTEGER, OUT p3 INTEGER)
            BEGIN
                SET p3 := p1 + p2;
            END;"""
        )

        tracer.enabled = True
        proc = "sp_sum"
        data = (40, 2, None)

        # spans[len(spans) - 2]
        cursor.callproc(proc, data)

        # spans[len(spans) - 1]
        cursor.execute(
            """
                       SELECT @_sp_sum_0, @_sp_sum_1, @_sp_sum_2
                       """
        )
        output = cursor.fetchone()
        assert len(output) == 3
        assert output[2] == 42

        spans = tracer.pop()
        assert spans, spans

        # number of spans depends on PyMySQL implementation details,
        # typically, internal calls to execute, but at least we
        # can expect the last closed span to be our proc.
        span = spans[len(spans) - 2]
        assert_is_measured(span)
        assert span.service == "pymysql"
        assert span.name == "pymysql.query"
        assert span.span_type == "sql"
        assert span.error == 0
        assert span.get_metric("out.port") == MYSQL_CONFIG.get("port")
        meta = {}
        meta.update(self.DB_INFO)
        assert_dict_issuperset(span.get_tags(), meta)

    def test_simple_query_ot(self):
        """OpenTracing version of test_simple_query."""
        conn, tracer = self._get_conn_tracer()

        ot_tracer = init_tracer("mysql_svc", tracer)
        with ot_tracer.start_active_span("mysql_op"):
            cursor = conn.cursor()
            cursor.execute("SELECT 1")
            rows = cursor.fetchall()
            assert len(rows) == 1

        spans = tracer.pop()
        assert len(spans) == 2
        ot_span, dd_span = spans

        # confirm parenting
        assert ot_span.parent_id is None
        assert dd_span.parent_id == ot_span.span_id

        assert ot_span.service == "mysql_svc"
        assert ot_span.name == "mysql_op"

        assert_is_measured(dd_span)
        assert dd_span.service == "pymysql"
        assert dd_span.name == "pymysql.query"
        assert dd_span.span_type == "sql"
        assert dd_span.error == 0
        assert dd_span.get_metric("out.port") == MYSQL_CONFIG.get("port")
        meta = {}
        meta.update(self.DB_INFO)
        assert_dict_issuperset(dd_span.get_tags(), meta)

    def test_simple_query_ot_fetchall(self):
        """OpenTracing version of test_simple_query."""
        with self.override_config("pymysql", dict(trace_fetch_methods=True)):
            conn, tracer = self._get_conn_tracer()

            ot_tracer = init_tracer("mysql_svc", tracer)
            with ot_tracer.start_active_span("mysql_op"):
                cursor = conn.cursor()
                cursor.execute("SELECT 1")
                rows = cursor.fetchall()
                assert len(rows) == 1

            spans = tracer.pop()
            assert len(spans) == 3
            ot_span, dd_span, fetch_span = spans

            # confirm parenting
            assert ot_span.parent_id is None
            assert dd_span.parent_id == ot_span.span_id

            assert ot_span.service == "mysql_svc"
            assert ot_span.name == "mysql_op"

            assert_is_measured(dd_span)
            assert dd_span.service == "pymysql"
            assert dd_span.name == "pymysql.query"
            assert dd_span.span_type == "sql"
            assert dd_span.error == 0
            assert dd_span.get_metric("out.port") == MYSQL_CONFIG.get("port")
            meta = {}
            meta.update(self.DB_INFO)
            assert_dict_issuperset(dd_span.get_tags(), meta)

            assert fetch_span.name == "pymysql.query.fetchall"

    def test_commit(self):
        conn, tracer = self._get_conn_tracer()

        conn.commit()
        spans = tracer.pop()
        assert len(spans) == 1
        span = spans[0]
        assert span.service == "pymysql"
        assert span.name == "pymysql.connection.commit"

    def test_rollback(self):
        conn, tracer = self._get_conn_tracer()

        conn.rollback()
        spans = tracer.pop()
        assert len(spans) == 1
        span = spans[0]
        assert span.service == "pymysql"
        assert span.name == "pymysql.connection.rollback"

    def test_analytics_default(self):
        conn, tracer = self._get_conn_tracer()

        cursor = conn.cursor()
        cursor.execute("SELECT 1")
        rows = cursor.fetchall()
        assert len(rows) == 1
        spans = tracer.pop()

        self.assertEqual(len(spans), 1)
        span = spans[0]
        self.assertIsNone(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY))

    def test_analytics_with_rate(self):
        with self.override_config("pymysql", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
            conn, tracer = self._get_conn_tracer()

            cursor = conn.cursor()
            cursor.execute("SELECT 1")
            rows = cursor.fetchall()
            assert len(rows) == 1
            spans = tracer.pop()

            self.assertEqual(len(spans), 1)
            span = spans[0]
            self.assertEqual(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 0.5)

    def test_analytics_without_rate(self):
        with self.override_config("pymysql", dict(analytics_enabled=True)):
            conn, tracer = self._get_conn_tracer()

            cursor = conn.cursor()
            cursor.execute("SELECT 1")
            rows = cursor.fetchall()
            assert len(rows) == 1
            spans = tracer.pop()

            self.assertEqual(len(spans), 1)
            span = spans[0]
            self.assertEqual(span.get_metric(ANALYTICS_SAMPLE_RATE_KEY), 1.0)


class TestPyMysqlPatch(PyMySQLCore, TracerTestCase):
    def _get_conn_tracer(self):
        if not self.conn:
            self.conn = pymysql.connect(**MYSQL_CONFIG)
            assert not self.conn._closed
            # Ensure that the default pin is there, with its default value
            pin = Pin.get_from(self.conn)
            assert pin
            # Customize the service
            # we have to apply it on the existing one since new one won't inherit `app`
            pin.clone(tracer=self.tracer).onto(self.conn)

            return self.conn, self.tracer

    def test_patch_unpatch(self):
        unpatch()
        # assert we start unpatched
        conn = pymysql.connect(**MYSQL_CONFIG)
        assert not Pin.get_from(conn)
        conn.close()

        patch()
        try:
            conn = pymysql.connect(**MYSQL_CONFIG)
            pin = Pin.get_from(conn)
            assert pin
            pin.clone(tracer=self.tracer).onto(conn)
            assert not conn._closed

            cursor = conn.cursor()
            cursor.execute("SELECT 1")
            rows = cursor.fetchall()
            assert len(rows) == 1
            spans = self.pop_spans()
            assert len(spans) == 1

            span = spans[0]
            assert span.service == "pymysql"
            assert span.name == "pymysql.query"
            assert span.span_type == "sql"
            assert span.error == 0
            assert span.get_metric("out.port") == MYSQL_CONFIG.get("port")

            meta = {}
            meta.update(self.DB_INFO)
            assert_dict_issuperset(span.get_tags(), meta)
        finally:
            unpatch()

            # assert we finish unpatched
            conn = pymysql.connect(**MYSQL_CONFIG)
            assert not Pin.get_from(conn)
            conn.close()

        patch()

    def test_user_pin_override(self):
        conn, tracer = self._get_conn_tracer()
        pin = Pin.get_from(conn)
        pin.clone(service="pin-svc", tracer=self.tracer).onto(conn)
        cursor = conn.cursor()
        cursor.execute("SELECT 1")
        rows = cursor.fetchall()
        assert len(rows) == 1
        spans = tracer.pop()
        assert len(spans) == 1

        span = spans[0]
        assert span.service == "pin-svc"

    def test_context_manager(self):
        conn, tracer = self._get_conn_tracer()
        # connection doesn't support context manager usage
        with conn.cursor() as cursor:
            cursor.execute("SELECT 1")
            rows = cursor.fetchall()
            assert len(rows) == 1
        spans = tracer.pop()
        assert len(spans) == 1

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_PYMYSQL_SERVICE="mysvc"))
    def test_user_specified_service_integration(self):
        conn, tracer = self._get_conn_tracer()

        conn.rollback()
        spans = self.pop_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.service == "mysvc"



"""
Instrument dogpile.cache__ to report all cached lookups.

This will add spans around the calls to your cache backend (e.g. redis, memory,
etc). The spans will also include the following tags:

- key/keys: The key(s) dogpile passed to your backend. Note that this will be
  the output of the region's ``function_key_generator``, but before any key
  mangling is applied (i.e. the region's ``key_mangler``).
- region: Name of the region.
- backend: Name of the backend class.
- hit: If the key was found in the cache.
- expired: If the key is expired. This is only relevant if the key was found.

While cache tracing will generally already have keys in tags, some caching
setups will not have useful tag values - such as when you're using consistent
hashing with memcached - the key(s) will appear as a mangled hash.
::

    # Patch before importing dogpile.cache
    from ddtrace import patch
    patch(dogpile_cache=True)

    from dogpile.cache import make_region

    region = make_region().configure(
        "dogpile.cache.pylibmc",
        expiration_time=3600,
        arguments={"url": ["127.0.0.1"]},
    )

    @region.cache_on_arguments()
    def hello(name):
        # Some complicated, slow calculation
        return "Hello, {}".format(name)

.. __: https://dogpilecache.sqlalchemy.org/
"""
from ...internal.utils.importlib import require_modules


required_modules = ["dogpile.cache"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch
        from .patch import unpatch

        __all__ = ["patch", "unpatch"]



import django
import pytest

from ddtrace.internal import _context
from ddtrace.internal.compat import urlencode
from tests.utils import assert_span_http_status_code
from tests.utils import override_global_config


@pytest.mark.skipif(django.VERSION < (1, 10), reason="requires django version >= 1.10")
@pytest.mark.django_db
def test_trace_exceptions(client, test_spans):  # noqa flake8 complains about shadowing test_spans
    response = client.get("/users/")

    # Our custom exception handler is setting the status code to 500
    assert response.status_code == 500

    sp = test_spans.get_root_span()
    assert sp.name == "django.request"
    if django.VERSION >= (2, 2, 0):
        assert sp.resource == "GET ^users/$"
    else:
        assert sp.resource == "GET tests.contrib.djangorestframework.app.views.UserViewSet"
    assert sp.error == 1
    assert sp.span_type == "web"
    assert_span_http_status_code(sp, 500)
    assert sp.get_tag("http.method") == "GET"

    # the DRF integration should set the traceback on the django.view.dispatch span
    # (as it's the current span when the exception info is set)
    view_dispatch_spans = list(test_spans.filter_spans(name="django.view.dispatch"))
    assert len(view_dispatch_spans) == 1
    err_span = view_dispatch_spans[0]
    assert err_span.error == 1
    assert err_span.get_tag("error.msg") == "Authentication credentials were not provided."
    assert "NotAuthenticated" in err_span.get_tag("error.stack")


@pytest.mark.skipif(django.VERSION < (1, 10), reason="requires django version >= 1.10")
def test_djangorest_request_body_urlencoded(client, test_spans, tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        tracer._appsec_enabled = True
        # Hack: need to pass an argument to configure so that the processors are recreated
        tracer.configure(api_version="v0.4")
        payload = urlencode({"mytestingbody_key": "mytestingbody_value"})
        client.post("/users/", payload, content_type="application/x-www-form-urlencoded")
        root_span = test_spans.spans[0]
        query = dict(_context.get_item("http.request.body", span=root_span))

        assert root_span.get_tag("_dd.appsec.json") is None
        assert query == {"mytestingbody_key": "mytestingbody_value"}



import os
import subprocess
import sys

import bm


class DDtraceRun(bm.Scenario):
    ddtrace_run = bm.var_bool()
    http = bm.var_bool()
    runtimemetrics = bm.var_bool()
    telemetry = bm.var_bool()
    profiling = bm.var_bool()
    appsec = bm.var_bool()
    tracing = bm.var_bool()

    def run(self):
        # setup subprocess environment variables
        env = os.environ.copy()
        env["DD_RUNTIME_METRICS_ENABLED"] = str(self.runtimemetrics)
        env["DD_APPSEC_ENABLED"] = str(self.appsec)

        # initialize subprocess args
        subp_cmd = []
        code = "import ddtrace; ddtrace.patch_all()\n"
        if self.ddtrace_run:
            subp_cmd = ["ddtrace-run"]
            code = ""

        if self.http:
            # mock requests to the trace agent before starting services
            env["DD_TRACE_API_VERSION"] = "v0.4"
            code += """
import httpretty
from ddtrace import tracer
from ddtrace.internal.telemetry import telemetry_writer

httpretty.enable(allow_net_connect=False)
httpretty.register_uri(httpretty.PUT, '%s/%s' % (tracer.agent_trace_url, 'v0.4/traces'))
httpretty.register_uri(httpretty.POST, '%s/%s' % (tracer.agent_trace_url, telemetry_writer.ENDPOINT))
# profiler will collect snapshot during shutdown
httpretty.register_uri(httpretty.POST, '%s/%s' % (tracer.agent_trace_url, 'profiling/v1/input'))
"""

        if self.telemetry:
            code += "telemetry_writer.enable()\n"

        if self.tracing:
            code += "span = tracer.trace('test-x', service='bench-test'); span.finish()\n"

        if self.profiling:
            code += "import ddtrace.profiling.auto\n"

        # stage code for execution in a subprocess
        subp_cmd += [sys.executable, "-c", code]

        def _(loops):
            for _ in range(loops):
                subprocess.check_output(subp_cmd, env=env)

        yield _



SERVICE = "kombu"

# net extension
VHOST = "out.vhost"

# standard tags
EXCHANGE = "kombu.exchange"
BODY_LEN = "kombu.body_length"
ROUTING_KEY = "kombu.routing_key"

PUBLISH_NAME = "kombu.publish"
RECEIVE_NAME = "kombu.receive"



"""
The pytest integration traces test executions.

Enabling
~~~~~~~~

Enable traced execution of tests using ``pytest`` runner by
running ``pytest --ddtrace`` or by modifying any configuration
file read by pytest (``pytest.ini``, ``setup.cfg``, ...)::

    [pytest]
    ddtrace = 1

You can enable all integrations by using the ``--ddtrace-patch-all`` option
alongside ``--ddtrace`` or by adding this to your configuration::

    [pytest]
    ddtrace = 1
    ddtrace-patch-all = 1


.. note::
   The ddtrace plugin for pytest has the side effect of importing the ddtrace
   package and starting a global tracer.

   If this is causing issues for your pytest runs where traced execution of
   tests is not enabled, you can deactivate the plugin::

     [pytest]
     addopts = -p no:ddtrace

   See the `pytest documentation
   <https://docs.pytest.org/en/7.1.x/how-to/plugins.html#deactivating-unregistering-a-plugin-by-name>`_
   for more details.


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.pytest["service"]

   The service name reported by default for pytest traces.

   This option can also be set with the integration specific ``DD_PYTEST_SERVICE`` environment
   variable, or more generally with the `DD_SERVICE` environment variable.

   Default: Name of the repository being tested, otherwise ``"pytest"`` if the repository name cannot be found.


.. py:data:: ddtrace.config.pytest["operation_name"]

   The operation name reported by default for pytest traces.

   This option can also be set with the ``DD_PYTEST_OPERATION_NAME`` environment
   variable.

   Default: ``"pytest.test"``
"""

import os

from ddtrace import config


# pytest default settings
config._add(
    "pytest",
    dict(
        _default_service="pytest",
        operation_name=os.getenv("DD_PYTEST_OPERATION_NAME", default="pytest.test"),
    ),
)



import typing
import weakref

def get_thread_name(thread_id: int) -> str: ...
def get_thread_native_id(thread_id: int) -> int: ...

_T = typing.TypeVar("_T")

class _ThreadLink(typing.Generic[_T]):
    def link_object(self, obj: _T) -> None: ...
    def clear_threads(self, existing_thread_ids: typing.Set[int]) -> None: ...
    def get_object(self, thread_id: int) -> typing.Optional[_T]: ...



"""
Trace queries to botocore api done via a pynamodb client
"""

import pynamodb.connection.base

from ddtrace import config
from ddtrace.vendor import wrapt

from .. import trace_utils
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...internal.utils import ArgumentError
from ...internal.utils import get_argument_value
from ...internal.utils.formats import deep_getattr
from ...pin import Pin
from ..trace_utils import unwrap


# Pynamodb connection class
_PynamoDB_client = pynamodb.connection.base.Connection

config._add(
    "pynamodb",
    {
        "_default_service": "pynamodb",
    },
)


def patch():
    if getattr(pynamodb.connection.base, "_datadog_patch", False):
        return
    setattr(pynamodb.connection.base, "_datadog_patch", True)

    wrapt.wrap_function_wrapper("pynamodb.connection.base", "Connection._make_api_call", patched_api_call)
    Pin(service=None).onto(pynamodb.connection.base.Connection)


def unpatch():
    if getattr(pynamodb.connection.base, "_datadog_patch", False):
        setattr(pynamodb.connection.base, "_datadog_patch", False)
        unwrap(pynamodb.connection.base.Connection, "_make_api_call")


def patched_api_call(original_func, instance, args, kwargs):

    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return original_func(*args, **kwargs)

    with pin.tracer.trace(
        "pynamodb.command", service=trace_utils.ext_service(pin, config.pynamodb, "pynamodb"), span_type=SpanTypes.HTTP
    ) as span:

        span.set_tag(SPAN_MEASURED_KEY)

        try:
            operation = get_argument_value(args, kwargs, 0, "operation_name")
            span.resource = operation

            if args[1] and "TableName" in args[1]:
                table_name = args[1]["TableName"]
                span.set_tag("table_name", table_name)
                span.resource = span.resource + " " + table_name

        except ArgumentError:
            span.resource = "Unknown"
            operation = None

        region_name = deep_getattr(instance, "client.meta.region_name")

        meta = {
            "aws.agent": "pynamodb",
            "aws.operation": operation,
            "aws.region": region_name,
        }
        span.set_tags(meta)

        # set analytics sample rate
        sample_rate = config.pynamodb.get_analytics_sample_rate(use_global_config=True)

        if sample_rate is not None:
            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)

        result = original_func(*args, **kwargs)

        return result



class SpanTypes(object):
    CACHE = "cache"
    CASSANDRA = "cassandra"
    ELASTICSEARCH = "elasticsearch"
    GRPC = "grpc"
    GRAPHQL = "graphql"
    HTTP = "http"
    MONGODB = "mongodb"
    REDIS = "redis"
    SQL = "sql"
    TEMPLATE = "template"
    TEST = "test"
    WEB = "web"
    WORKER = "worker"



import aiohttp_jinja2
import pytest

from ddtrace.contrib.aiohttp_jinja2 import patch
from ddtrace.contrib.aiohttp_jinja2 import unpatch
from ddtrace.pin import Pin
from tests.contrib.aiohttp.conftest import app_tracer  # noqa
from tests.contrib.aiohttp.conftest import patched_app_tracer  # noqa
from tests.contrib.aiohttp.conftest import untraced_app_tracer  # noqa


@pytest.fixture
def patched_app_tracer_jinja(patched_app_tracer):  # noqa: F811
    app, tracer = patched_app_tracer
    patch()
    Pin.override(aiohttp_jinja2, tracer=tracer)
    yield app, tracer
    unpatch()


@pytest.fixture
def untraced_app_tracer_jinja(untraced_app_tracer):  # noqa: F811
    patch()
    app, tracer = untraced_app_tracer
    Pin.override(aiohttp_jinja2, tracer=tracer)
    yield app, tracer
    unpatch()



import pytest
import tornado

from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import ORIGIN_KEY
from ddtrace.constants import SAMPLING_PRIORITY_KEY
from ddtrace.ext import http
from tests.opentracer.utils import init_tracer
from tests.utils import assert_is_measured
from tests.utils import assert_span_http_status_code

from .utils import TornadoTestCase
from .utils import TracerTestCase
from .web.app import CustomDefaultHandler


class TestTornadoWeb(TornadoTestCase):
    """
    Ensure that Tornado web handlers are properly traced.
    """

    def test_success_handler(self, query_string=""):
        # it should trace a handler that returns 200
        if query_string:
            fqs = "?" + query_string
        else:
            fqs = ""
        response = self.fetch("/success/" + fqs)
        assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.SuccessHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 200)
        if config.tornado.trace_query_string:
            assert self.get_url("/success/") + fqs == request_span.get_tag(http.URL)
            assert query_string == request_span.get_tag(http.QUERY_STRING)
        else:
            assert self.get_url("/success/") == request_span.get_tag(http.URL)
            assert http.QUERY_STRING not in request_span.get_tags()
        assert 0 == request_span.error

    def test_success_handler_query_string(self):
        self.test_success_handler("foo=bar")

    def test_success_handler_query_string_trace(self):
        with self.override_http_config("tornado", dict(trace_query_string=True)):
            self.test_success_handler("foo=bar")

    def test_status_code_500_handler(self):
        """
        Test an endpoint which sets the status code to 500 but doesn't raise an exception

        We expect the resulting span to be marked as an error
        """
        response = self.fetch("/status_code/500")
        assert 500 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tests.contrib.tornado.web.app.ResponseStatusHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 500)
        assert self.get_url("/status_code/500") == request_span.get_tag(http.URL)
        assert 1 == request_span.error

    def test_nested_handler(self):
        # it should trace a handler that calls the tracer.trace() method
        # using the automatic Context retrieval
        response = self.fetch("/nested/")
        assert 200 == response.code
        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 2 == len(traces[0])
        # check request span
        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.NestedHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/nested/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error
        # check nested span
        nested_span = traces[0][1]
        assert "tornado-web" == nested_span.service
        assert "tornado.sleep" == nested_span.name
        assert 0 == nested_span.error
        # check durations because of the yield sleep
        assert request_span.duration >= 0.05
        assert nested_span.duration >= 0.05

    def test_exception_handler(self):
        # it should trace a handler that raises an exception
        response = self.fetch("/exception/")
        assert 500 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.ExceptionHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 500)
        assert self.get_url("/exception/") == request_span.get_tag(http.URL)
        assert 1 == request_span.error
        assert "Ouch!" == request_span.get_tag("error.msg")
        assert "Exception: Ouch!" in request_span.get_tag("error.stack")

    def test_http_exception_handler(self):
        # it should trace a handler that raises a Tornado HTTPError
        response = self.fetch("/http_exception/")
        assert 501 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.HTTPExceptionHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 501)
        assert self.get_url("/http_exception/") == request_span.get_tag(http.URL)
        assert 1 == request_span.error
        assert "HTTP 501: Not Implemented (unavailable)" == request_span.get_tag("error.msg")
        assert "HTTP 501: Not Implemented (unavailable)" in request_span.get_tag("error.stack")

    def test_http_exception_500_handler(self):
        # it should trace a handler that raises a Tornado HTTPError
        response = self.fetch("/http_exception_500/")
        assert 500 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.HTTPException500Handler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 500)
        assert self.get_url("/http_exception_500/") == request_span.get_tag(http.URL)
        assert 1 == request_span.error
        assert "HTTP 500: Server Error (server error)" == request_span.get_tag("error.msg")
        assert "HTTP 500: Server Error (server error)" in request_span.get_tag("error.stack")

    def test_sync_success_handler(self):
        # it should trace a synchronous handler that returns 200
        response = self.fetch("/sync_success/")
        assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.SyncSuccessHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/sync_success/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

    def test_sync_exception_handler(self):
        # it should trace a handler that raises an exception
        response = self.fetch("/sync_exception/")
        assert 500 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.SyncExceptionHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 500)
        assert self.get_url("/sync_exception/") == request_span.get_tag(http.URL)
        assert 1 == request_span.error
        assert "Ouch!" == request_span.get_tag("error.msg")
        assert "Exception: Ouch!" in request_span.get_tag("error.stack")

    def test_404_handler(self):
        # it should trace 404
        response = self.fetch("/does_not_exist/")
        assert 404 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tornado.web.ErrorHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 404)
        assert self.get_url("/does_not_exist/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

    def test_redirect_handler(self):
        # it should trace the built-in RedirectHandler
        response = self.fetch("/redirect/")
        assert 200 == response.code

        # we trace two different calls: the RedirectHandler and the SuccessHandler
        traces = self.pop_traces()
        assert 2 == len(traces)
        assert 1 == len(traces[0])
        assert 1 == len(traces[1])

        redirect_span = traces[0][0]
        assert_is_measured(redirect_span)
        assert "tornado-web" == redirect_span.service
        assert "tornado.request" == redirect_span.name
        assert "web" == redirect_span.span_type
        assert "tornado.web.RedirectHandler" == redirect_span.resource
        assert "GET" == redirect_span.get_tag("http.method")
        assert_span_http_status_code(redirect_span, 301)
        assert self.get_url("/redirect/") == redirect_span.get_tag(http.URL)
        assert 0 == redirect_span.error

        success_span = traces[1][0]
        assert "tornado-web" == success_span.service
        assert "tornado.request" == success_span.name
        assert "web" == success_span.span_type
        assert "tests.contrib.tornado.web.app.SuccessHandler" == success_span.resource
        assert "GET" == success_span.get_tag("http.method")
        assert_span_http_status_code(success_span, 200)
        assert self.get_url("/success/") == success_span.get_tag(http.URL)
        assert 0 == success_span.error

    def test_static_handler(self):
        # it should trace the access to static files
        response = self.fetch("/statics/empty.txt")
        assert 200 == response.code
        assert "Static file\n" == response.body.decode("utf-8")

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert_is_measured(request_span)
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tornado.web.StaticFileHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/statics/empty.txt") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

    def test_propagation(self):
        # it should trace a handler that returns 200 with a propagated context
        headers = {"x-datadog-trace-id": "1234", "x-datadog-parent-id": "4567", "x-datadog-sampling-priority": "2"}
        response = self.fetch("/success/", headers=headers)
        assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]

        # simple sanity check on the span
        assert "tornado.request" == request_span.name
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/success/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

        # check propagation
        assert 1234 == request_span.trace_id
        assert 4567 == request_span.parent_id
        assert 2 == request_span.get_metric(SAMPLING_PRIORITY_KEY)

    # Opentracing support depends on new AsyncioScopeManager
    # See: https://github.com/opentracing/opentracing-python/pull/118
    @pytest.mark.skipif(
        tornado.version_info >= (5, 0), reason="Opentracing ScopeManager not available for Tornado >= 5"
    )
    def test_success_handler_ot(self):
        """OpenTracing version of test_success_handler."""
        from opentracing.scope_managers.tornado import TornadoScopeManager

        ot_tracer = init_tracer("tornado_svc", self.tracer, scope_manager=TornadoScopeManager())

        with ot_tracer.start_active_span("tornado_op"):
            response = self.fetch("/success/")
            assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 2 == len(traces[0])
        # dd_span will start and stop before the ot_span finishes
        ot_span, dd_span = traces[0]

        # confirm the parenting
        assert ot_span.parent_id is None
        assert dd_span.parent_id == ot_span.span_id

        assert ot_span.name == "tornado_op"
        assert ot_span.service == "tornado_svc"

        assert_is_measured(dd_span)
        assert "tornado-web" == dd_span.service
        assert "tornado.request" == dd_span.name
        assert "web" == dd_span.span_type
        assert "tests.contrib.tornado.web.app.SuccessHandler" == dd_span.resource
        assert "GET" == dd_span.get_tag("http.method")
        assert_span_http_status_code(dd_span, 200)
        assert self.get_url("/success/") == dd_span.get_tag(http.URL)
        assert 0 == dd_span.error


class TestTornadoWebAnalyticsDefault(TornadoTestCase):
    """
    Ensure that Tornado web handlers generate APM events with default settings
    """

    def test_analytics_global_on_integration_default(self):
        """
        When making a request
            When an integration trace search is not event sample rate is not set and globally trace search is enabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=True)):
            # it should trace a handler that returns 200
            response = self.fetch("/success/")
            self.assertEqual(200, response.code)

            self.assert_structure(
                dict(name="tornado.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 1.0}),
            )

    def test_analytics_global_off_integration_default(self):
        """
        When making a request
            When an integration trace search is not set and sample rate is set and globally trace search is disabled
                We expect the root span to not include tag
        """
        with self.override_global_config(dict(analytics_enabled=False)):
            # it should trace a handler that returns 200
            response = self.fetch("/success/")
            self.assertEqual(200, response.code)

            root = self.get_root_span()
            self.assertIsNone(root.get_metric(ANALYTICS_SAMPLE_RATE_KEY))


class TestTornadoWebAnalyticsOn(TornadoTestCase):
    """
    Ensure that Tornado web handlers generate APM events with default settings
    """

    def get_settings(self):
        # distributed_tracing needs to be disabled manually
        return {
            "datadog_trace": {
                "analytics_enabled": True,
                "analytics_sample_rate": 0.5,
            },
        }

    def test_analytics_global_on_integration_on(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set and globally trace search is enabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=True)):
            # it should trace a handler that returns 200
            response = self.fetch("/success/")
            self.assertEqual(200, response.code)

            self.assert_structure(
                dict(name="tornado.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 0.5}),
            )

    def test_analytics_global_off_integration_on(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set and globally trace search is disabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=False)):
            # it should trace a handler that returns 200
            response = self.fetch("/success/")
            self.assertEqual(200, response.code)

            self.assert_structure(
                dict(name="tornado.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 0.5}),
            )


class TestTornadoWebAnalyticsNoRate(TornadoTestCase):
    """
    Ensure that Tornado web handlers generate APM events with default settings
    """

    def get_settings(self):
        # distributed_tracing needs to be disabled manually
        return {
            "datadog_trace": {
                "analytics_enabled": True,
            },
        }

    def test_analytics_global_on_integration_on(self):
        """
        When making a request
            When an integration trace search is enabled and sample rate is set and globally trace search is enabled
                We expect the root span to have the appropriate tag
        """
        with self.override_global_config(dict(analytics_enabled=True)):
            # it should trace a handler that returns 200
            response = self.fetch("/success/")
            self.assertEqual(200, response.code)

            self.assert_structure(
                dict(name="tornado.request", metrics={ANALYTICS_SAMPLE_RATE_KEY: 1.0}),
            )


class TestNoPropagationTornadoWebViaSetting(TornadoTestCase):
    """
    Ensure that Tornado web handlers are properly traced and are ignoring propagated HTTP headers when disabled.
    """

    def get_settings(self):
        # distributed_tracing needs to be disabled manually
        return {
            "datadog_trace": {
                "distributed_tracing": False,
            },
        }

    def test_no_propagation(self):
        # it should not propagate the HTTP context
        headers = {
            "x-datadog-trace-id": "1234",
            "x-datadog-parent-id": "4567",
            "x-datadog-sampling-priority": "2",
            "x-datadog-origin": "synthetics",
        }
        response = self.fetch("/success/", headers=headers)
        assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]

        # simple sanity check on the span
        assert "tornado.request" == request_span.name
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/success/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

        # check non-propagation
        assert request_span.trace_id != 1234
        assert request_span.parent_id != 4567
        assert request_span.get_metric(SAMPLING_PRIORITY_KEY) != 2
        assert request_span.get_tag(ORIGIN_KEY) != "synthetics"


class TestNoPropagationTornadoWebViaConfig(TornadoTestCase):
    """
    Ensure that Tornado web handlers are properly traced and are ignoring propagated HTTP headers when disabled.
    """

    def test_no_propagation_via_int_config(self):
        config.tornado.distributed_tracing = False
        # it should not propagate the HTTP context
        headers = {
            "x-datadog-trace-id": "1234",
            "x-datadog-parent-id": "4567",
            "x-datadog-sampling-priority": "2",
            "x-datadog-origin": "synthetics",
        }
        response = self.fetch("/success/", headers=headers)
        assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]

        # simple sanity check on the span
        assert "tornado.request" == request_span.name
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/success/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

        # check non-propagation
        assert request_span.trace_id != 1234
        assert request_span.parent_id != 4567
        assert request_span.get_metric(SAMPLING_PRIORITY_KEY) != 2
        assert request_span.get_tag(ORIGIN_KEY) != "synthetics"

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_TORNADO_DISTRIBUTED_TRACING="False"))
    def test_no_propagation_via_env_var(self):
        # it should not propagate the HTTP context
        headers = {
            "x-datadog-trace-id": "1234",
            "x-datadog-parent-id": "4567",
            "x-datadog-sampling-priority": "2",
            "x-datadog-origin": "synthetics",
        }
        response = self.fetch("/success/", headers=headers)
        assert 200 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]

        # simple sanity check on the span
        assert "tornado.request" == request_span.name
        assert_span_http_status_code(request_span, 200)
        assert self.get_url("/success/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error

        # check non-propagation
        assert request_span.trace_id != 1234
        assert request_span.parent_id != 4567
        assert request_span.get_metric(SAMPLING_PRIORITY_KEY) != 2
        assert request_span.get_tag(ORIGIN_KEY) != "synthetics"


class TestCustomTornadoWeb(TornadoTestCase):
    """
    Ensure that Tornado web handlers are properly traced when using
    a custom default handler.
    """

    def get_settings(self):
        return {
            "default_handler_class": CustomDefaultHandler,
            "default_handler_args": dict(status_code=400),
        }

    def test_custom_default_handler(self):
        # it should trace any call that uses a custom default handler
        response = self.fetch("/custom_handler/")
        assert 400 == response.code

        traces = self.pop_traces()
        assert 1 == len(traces)
        assert 1 == len(traces[0])

        request_span = traces[0][0]
        assert "tornado-web" == request_span.service
        assert "tornado.request" == request_span.name
        assert "web" == request_span.span_type
        assert "tests.contrib.tornado.web.app.CustomDefaultHandler" == request_span.resource
        assert "GET" == request_span.get_tag("http.method")
        assert_span_http_status_code(request_span, 400)
        assert self.get_url("/custom_handler/") == request_span.get_tag(http.URL)
        assert 0 == request_span.error



import os

import tornado

import ddtrace
from ddtrace import config
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

from . import application
from . import compat
from . import context_provider
from . import decorators
from . import handlers
from . import template
from ...internal.utils.formats import asbool
from ...internal.utils.wrappers import unwrap as _u


config._add(
    "tornado",
    dict(
        distributed_tracing=asbool(os.getenv("DD_TORNADO_DISTRIBUTED_TRACING", default=True)),
    ),
)


def patch():
    """
    Tracing function that patches the Tornado web application so that it will be
    traced using the given ``tracer``.
    """
    # patch only once
    if getattr(tornado, "__datadog_patch", False):
        return
    setattr(tornado, "__datadog_patch", True)

    # patch Application to initialize properly our settings and tracer
    _w("tornado.web", "Application.__init__", application.tracer_config)

    # patch RequestHandler to trace all Tornado handlers
    _w("tornado.web", "RequestHandler._execute", handlers.execute)
    _w("tornado.web", "RequestHandler.on_finish", handlers.on_finish)
    _w("tornado.web", "RequestHandler.log_exception", handlers.log_exception)

    # patch Template system
    _w("tornado.template", "Template.generate", template.generate)

    # patch Python Futures if available when an Executor pool is used
    compat.wrap_futures()

    # configure the global tracer
    ddtrace.tracer.configure(
        context_provider=context_provider,
        wrap_executor=decorators.wrap_executor,
    )


def unpatch():
    """
    Remove all tracing functions in a Tornado web application.
    """
    if not getattr(tornado, "__datadog_patch", False):
        return
    setattr(tornado, "__datadog_patch", False)

    # unpatch Tornado
    _u(tornado.web.RequestHandler, "_execute")
    _u(tornado.web.RequestHandler, "on_finish")
    _u(tornado.web.RequestHandler, "log_exception")
    _u(tornado.web.Application, "__init__")
    _u(tornado.template.Template, "generate")

    # unpatch `futures`
    compat.unwrap_futures()
